@@ -2,18 +2,18 @@
 PC-BASIC - expressions.py
 Expression stack
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 from collections import deque
 from functools import partial
 import logging
-import string
 import struct
 import types
 
 from ..base import tokens as tk
+from ..base.tokens import DIGITS, LETTERS
 from ..base import error
 from .. import values
 from .. import dos
@@ -21,11 +21,6 @@ from . import operators as op
 from . import userfunctions
 
 
-# bytes constants
-DIGITS = string.digits
-LETTERS = string.ascii_letters
-
-
 class ExpressionParser(object):
     """Expression parser."""
 
@@ -137,7 +132,7 @@ class ExpressionParser(object):
             tk.FN: None,
             b'_': self._gen_parse_call_extension,
         }
-        self._functions = set(self._complex.keys() + self._simple.keys())
+        self._functions = set(self._complex.keys()) | set(self._simple.keys())
 
     def init_functions(self, session):
         """Initialise function callbacks."""
@@ -160,14 +155,14 @@ class ExpressionParser(object):
             tk.ERDEV + b'$': session.files.erdev_str_,
             tk.VARPTR: session.memory.varptr_,
             tk.VARPTR + b'$': session.memory.varptr_str_,
-            tk.SCREEN: session.screen.screen_fn_,
+            tk.SCREEN: session.text_screen.screen_fn_,
             tk.FN: None,
             tk.ERL: session.interpreter.erl_,
             tk.ERR: session.interpreter.err_,
             tk.STRING: values.string_,
             tk.INSTR: values.instr_,
-            tk.CSRLIN: session.screen.csrlin_,
-            tk.POINT: session.drawing.point_,
+            tk.CSRLIN: session.text_screen.csrlin_,
+            tk.POINT: session.graphics.point_,
             tk.INKEY: session.keyboard.inkey_,
             tk.CVI: values.cvi_,
             tk.CVS: values.cvs_,
@@ -180,7 +175,7 @@ class ExpressionParser(object):
             tk.TIME: session.clock.time_fn_,
             tk.PLAY: session.sound.play_fn_,
             tk.TIMER: session.clock.timer_,
-            tk.PMAP: session.drawing.pmap_,
+            tk.PMAP: session.graphics.pmap_,
             tk.LEFT: values.left_,
             tk.RIGHT: values.right_,
             tk.MID: values.mid_,
@@ -197,7 +192,7 @@ class ExpressionParser(object):
             tk.ATN: values.atn_,
             tk.FRE: session.memory.fre_,
             tk.INP: session.machine.inp_,
-            tk.POS: session.screen.pos_,
+            tk.POS: session.text_screen.pos_,
             tk.LEN: values.len_,
             tk.STR: values.str_,
             tk.VAL: values.val_,
@@ -235,7 +230,7 @@ class ExpressionParser(object):
         self.__dict__.update(pickle_dict)
         self._init_syntax()
 
-    def parse_expression(self,ins):
+    def parse_expression(self, ins):
         """Parse and evaluate tokenised expression."""
         self._memory.strings.reset_temporaries()
         return self.parse(ins)
@@ -257,7 +252,6 @@ class ExpressionParser(object):
                     break
                 elif d in op.OPERATORS:
                     ins.read(len(d))
-                    prec = op.PRECEDENCE[d]
                     # get combined operators such as >=
                     if d in op.COMBINABLE:
                         nxt = ins.skip_blank()
@@ -270,12 +264,14 @@ class ExpressionParser(object):
                         # because it will be seen as an illegal unary
                         try:
                             oper = op.UNARY[d]
+                            prec = op.PRECEDENCE[(d, nargs)]
                         except KeyError:
                             raise error.BASICError(error.STX)
                     else:
                         nargs = 2
                         try:
                             oper = op.BINARY[d]
+                            prec = op.PRECEDENCE[(d, nargs)]
                         except KeyError:
                             # illegal combined ops like == raise syntax error here
                             raise error.BASICError(error.STX)
@@ -413,7 +409,7 @@ class ExpressionParser(object):
     def _no_argument(self, ins):
         """No arguments to parse."""
         return
-        yield
+        yield # pragma: no cover
 
     def _gen_parse_arguments(self, ins, length=1):
         """Parse a comma-separated list of arguments."""
@@ -422,7 +418,7 @@ class ExpressionParser(object):
         ins.require_read((b'(',))
         for i in range(length-1):
             yield self.parse(ins)
-            ins.require_read((b','),)
+            ins.require_read((b',',))
         yield self.parse(ins)
         ins.require_read((b')',))
 
@@ -431,9 +427,9 @@ class ExpressionParser(object):
         ins.require_read((b'(',))
         yield self.parse(ins)
         for _ in range(length-2):
-            ins.require_read((b','),)
+            ins.require_read((b',',))
             yield self.parse(ins)
-        if ins.skip_blank_read_if((b',',),):
+        if ins.skip_blank_read_if((b',',)):
             yield self.parse(ins)
         else:
             yield None
