@@ -2,25 +2,22 @@
 PC-BASIC - values.py
 Types, values and conversions
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 import math
-import string
 import struct
 import functools
 
+from ...compat import int2byte
+
 from ..base import error
 from ..base import tokens as tk
 from . import numbers
 from . import strings
 
 
-# mark bytes conversion explicitly
-int2byte = chr
-
-
 # BASIC type sigils:
 # Integer (%) - stored as two's complement, little-endian
 # Single (!) - stored as 4-byte Microsoft Binary Format
@@ -50,10 +47,14 @@ TYPE_TO_CLASS = {
     DBL: numbers.Double
 }
 
+# cutoff for trigonometric functions
+# above this machine precision makes the result useless and machine/os dependent
+# this is close to what gw uses but not quite equivalent
+TRIG_MAX = 5e16
 
 def size_bytes(name):
     """Return the size of a value type, by variable name or type char."""
-    return TYPE_TO_SIZE[name[-1]]
+    return TYPE_TO_SIZE[name[-1:]]
 
 ###############################################################################
 # type checks
@@ -80,6 +81,9 @@ def pass_number(inp, err=error.TYPE_MISMATCH):
 def next_string(args):
     """Retrieve a string from an iterator and return as Python value."""
     expr = next(args)
+    return to_string_or_none(expr)
+
+def to_string_or_none(expr):
     if isinstance(expr, strings.String):
         return expr.to_value()
     elif expr is None:
@@ -125,13 +129,18 @@ def _call_float_function(fn, *args):
     feh = args[0].error_handler
     try:
         # to_float can overflow on Double.pos_max
-        args = [arg.to_float(arg._values.double_math) for arg in args]
+        args = [_arg.to_float(values.double_math) for _arg in args]
         floatcls = args[0].__class__
-        args = [arg.to_value() for arg in args]
-        return floatcls(None, values).from_value(fn(*args))
+        python_args = [_arg.to_value() for _arg in args]
+        result = fn(*python_args)
+        # python3 may return complex values for some real functions
+        # where python2 simply raises an error
+        if isinstance(result, complex):
+            raise ValueError('Non-real result')
+        return floatcls(None, values).from_value(result)
     except (ValueError, ArithmeticError) as e:
         # create positive infinity of the appropriate class
-        if arg._values.double_math and isinstance(args[0], numbers.Double):
+        if values.double_math and isinstance(args[0], numbers.Double):
             floatcls = numbers.Double
         else:
             floatcls = numbers.Single
@@ -146,9 +155,9 @@ class FloatErrorHandler(object):
     # types of errors that do not always interrupt execution
     soft_types = (error.OVERFLOW, error.DIVISION_BY_ZERO)
 
-    def __init__(self, screen):
+    def __init__(self, console):
         """Setup handler."""
-        self._screen = screen
+        self._console = console
         self._do_raise = False
 
     def suspend(self, do_raise):
@@ -164,25 +173,23 @@ class FloatErrorHandler(object):
             math_error = error.OVERFLOW
         elif isinstance(e, ZeroDivisionError):
             math_error = error.DIVISION_BY_ZERO
-        else:
+        else: # pragma: no cover
+            # shouldn't happen, we're only called with ValueError/ArithmeticError
             raise e
-        if (self._do_raise or self._screen is None or
-                math_error not in self.soft_types):
+        if (self._do_raise or self._console is None or math_error not in self.soft_types):
             # also raises exception in error_handle_mode!
             # in that case, prints a normal error message
             raise error.BASICError(math_error)
         else:
             # write a message & continue as normal
             # message should not include line number or trailing \xFF
-            self._screen.write_line(error.BASICError(math_error).message)
+            self._console.write_line(error.BASICError(math_error).message)
         # return max value for the appropriate float type
-        if e.args and e.args[0]:
-            if isinstance(e.args[0], numbers.Float):
-                return e.args[0]
-            elif isinstance(e.args[0], numbers.Integer):
-                # integer values are not soft-handled
-                raise error.BASICError(math_error)
-        return numbers.Single(None, self).from_bytes(numbers.Single.pos_max)
+        # integer operations should just raise the BASICError directly, they are not handled
+        if e.args and isinstance(e.args[0], numbers.Float):
+            return e.args[0]
+        else: # pragma: no cover
+            return numbers.Single(None, self).from_bytes(numbers.Single.pos_max)
 
 
 ###############################################################################
@@ -195,9 +202,10 @@ class Values(object):
         self.stringspace = string_space
         # double-precision EXP, SIN, COS, TAN, ATN, LOG
         self.double_math = double_math
+        self.error_handler = None
 
     def set_handler(self, handler):
-        """Initialise the error message screen."""
+        """Initialise the error message console."""
         self.error_handler = handler
 
     def create(self, buf):
@@ -256,7 +264,7 @@ class Values(object):
         """Convert number token to new Number temporary"""
         if not token:
             raise ValueError('Token must not be empty')
-        lead = bytes(token)[0]
+        lead = bytes(token)[0:1]
         if lead == tk.T_SINGLE:
             return numbers.Single(None, self).from_token(token)
         elif lead == tk.T_DOUBLE:
@@ -281,7 +289,7 @@ class Values(object):
         if word[:2] == b'&H':
             return self.new_integer().from_hex(word[2:])
         elif word[:1] == b'&':
-            return self.new_integer().from_oct(word[2:] if word[1:2] == 'O' else word[1:])
+            return self.new_integer().from_oct(word[2:] if word[1:2] == b'O' else word[1:])
         # we need to try to convert to int first,
         # mainly so that the tokeniser can output the right token type
         try:
@@ -290,7 +298,8 @@ class Values(object):
             # non-integer characters, try a float
             pass
         except error.BASICError as e:
-            if e.err != error.OVERFLOW:
+            if e.err != error.OVERFLOW: # pragma: no cover
+                # shouldn't happen, from_str only raises Overflow
                 raise
         # if allow_nonnum == False, raises ValueError for non-numerical characters
         is_double, mantissa, exp10 = numbers.str_to_decimal(word, allow_nonnum)
@@ -299,12 +308,6 @@ class Values(object):
         return self.new_single().from_decimal(mantissa, exp10)
 
 
-@float_safe
-def round(x):
-    """Round to nearest whole number without converting to int."""
-    return pass_number(x).to_float().iround()
-
-
 ###############################################################################
 # conversions
 
@@ -364,17 +367,17 @@ def to_int(inp, unsigned=False):
 def mki_(args):
     """MKI$: return the byte representation of an int."""
     x, = args
-    return x._values.new_string().from_str(to_integer(x).to_bytes())
+    return x._values.new_string().from_str(bytes(to_integer(x).to_bytes()))
 
 def mks_(args):
     """MKS$: return the byte representation of a single."""
     x, = args
-    return x._values.new_string().from_str(to_single(x).to_bytes())
+    return x._values.new_string().from_str(bytes(to_single(x).to_bytes()))
 
 def mkd_(args):
     """MKD$: return the byte representation of a double."""
     x, = args
-    return x._values.new_string().from_str(to_double(x).to_bytes())
+    return x._values.new_string().from_str(bytes(to_double(x).to_bytes()))
 
 def cvi_(args):
     """CVI: return the int value of a byte representation."""
@@ -525,17 +528,17 @@ def exp_(args):
 def sin_(args):
     """Sine."""
     x, = args
-    return _call_float_function(math.sin, x)
+    return _call_float_function(lambda _x: math.sin(_x) if abs(_x) < TRIG_MAX else 0., x)
 
 def cos_(args):
     """Cosine."""
     x, = args
-    return _call_float_function(math.cos, x)
+    return _call_float_function(lambda _x: math.cos(_x) if abs(_x) < TRIG_MAX else 1., x)
 
 def tan_(args):
     """Tangent."""
     x, = args
-    return _call_float_function(math.tan, x)
+    return _call_float_function(lambda _x: math.tan(_x) if abs(_x) < TRIG_MAX else 0., x)
 
 def atn_(args):
     """Inverse tangent."""
@@ -552,7 +555,7 @@ def log_(args):
 # string representations and characteristics
 
 def to_repr(inp, leading_space, type_sign):
-    """Convert BASIC number to Python str representation."""
+    """Convert BASIC number to Python bytes representation."""
     # PRINT, STR$ - yes leading space, no type sign
     # WRITE - no leading space, no type sign
     # LIST - no loading space, yes type sign
@@ -689,7 +692,7 @@ def string_(args):
         error.range_check(0, 255, asc_value_or_char.to_int())
     list(args)
     if isinstance(asc_value_or_char, strings.String):
-        char = asc_value_or_char.to_str()[0]
+        char = asc_value_or_char.to_str()[:1]
     else:
         # overflow if outside Integer range
         ascval = asc_value_or_char.to_integer().to_int()
@@ -703,8 +706,11 @@ def string_(args):
 @float_safe
 def pow(left, right):
     """Left^right."""
+    if isinstance(left, strings.String) or isinstance(right, strings.String):
+        raise error.BASICError(error.TYPE_MISMATCH)
     if left._values.double_math and (
-            isinstance(left, numbers.Double) or isinstance(right, numbers.Double)):
+            isinstance(left, numbers.Double) or isinstance(right, numbers.Double)
+        ):
         return _call_float_function(lambda a, b: a**b, to_double(left), to_double(right))
     elif isinstance(right, numbers.Integer):
         return left.to_single().clone().ipow_int(right)
