@@ -2,7 +2,7 @@
 PC-BASIC - interface.video
 Base class for video plugins
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -23,24 +23,14 @@ class VideoPlugin(object):
         self._video_queue = video_queue
         self._handlers = {
             signals.VIDEO_SET_MODE: self.set_mode,
-            signals.VIDEO_PUT_GLYPH: self.put_glyph,
+            signals.VIDEO_UPDATE: self.update,
             signals.VIDEO_CLEAR_ROWS: self.clear_rows,
-            signals.VIDEO_SCROLL_UP: self.scroll_up,
-            signals.VIDEO_SCROLL_DOWN: self.scroll_down,
+            signals.VIDEO_SCROLL: self.scroll,
             signals.VIDEO_SET_PALETTE: self.set_palette,
             signals.VIDEO_SET_CURSOR_SHAPE: self.set_cursor_shape,
-            signals.VIDEO_SET_CURSOR_ATTR: self.set_cursor_attr,
             signals.VIDEO_SHOW_CURSOR: self.show_cursor,
             signals.VIDEO_MOVE_CURSOR: self.move_cursor,
-            signals.VIDEO_SET_PAGE: self.set_page,
-            signals.VIDEO_COPY_PAGE: self.copy_page,
             signals.VIDEO_SET_BORDER_ATTR: self.set_border_attr,
-            signals.VIDEO_BUILD_GLYPHS: self.build_glyphs,
-            signals.VIDEO_PUT_PIXEL: self.put_pixel,
-            signals.VIDEO_PUT_INTERVAL: self.put_interval,
-            signals.VIDEO_FILL_INTERVAL: self.fill_interval,
-            signals.VIDEO_PUT_RECT: self.put_rect,
-            signals.VIDEO_FILL_RECT: self.fill_rect,
             signals.VIDEO_SET_CAPTION: self.set_caption_message,
             signals.VIDEO_SET_CLIPBOARD_TEXT: self.set_clipboard_text,
         }
@@ -70,7 +60,7 @@ class VideoPlugin(object):
                 return True
             # putting task_done before the execution avoids hanging on join() after an exception
             self._video_queue.task_done()
-            if signal.event_type == signals.VIDEO_QUIT:
+            if signal.event_type == signals.QUIT:
                 # close thread
                 self.alive = False
             else:
@@ -96,65 +86,35 @@ class VideoPlugin(object):
 
     # signal handlers
 
-    def set_mode(self, mode_info):
+    def set_mode(self, canvas_height, canvas_width, text_height, text_width):
         """Initialise a given text or graphics mode."""
 
     def set_caption_message(self, msg):
         """Add a message to the window caption."""
 
-    def set_clipboard_text(self, text, mouse):
+    def set_clipboard_text(self, text):
         """Put text on the clipboard."""
 
-    def set_palette(self, rgb_palette_0, rgb_palette_1, pack_pixels):
+    def set_palette(self, attributes, pack_pixels):
         """Build the palette."""
 
     def set_border_attr(self, attr):
         """Change the border attribute."""
 
-    def clear_rows(self, back_attr, start, stop):
+    def clear_rows(self, back_attr, start_row, stop_row):
         """Clear a range of screen rows."""
 
-    def set_page(self, vpage, apage):
-        """Set the visible and active page."""
-
-    def copy_page(self, src, dst):
-        """Copy source to destination page."""
-
-    def show_cursor(self, cursor_on):
+    def show_cursor(self, cursor_on, cursor_blinks):
         """Change visibility of cursor."""
 
-    def move_cursor(self, crow, ccol):
-        """Move the cursor to a new position."""
+    def move_cursor(self, row, col, attr, width):
+        """Move the cursor to a new position and set attribute and width."""
 
-    def set_cursor_attr(self, attr):
-        """Change attribute of cursor."""
+    def scroll(self, direction, start_row, stop_row, back_attr):
+        """Scroll the screen between start_row and stop_row. direction 1 is down, -1 up."""
 
-    def scroll_up(self, from_line, scroll_height, back_attr):
-        """Scroll the screen up between from_line and scroll_height."""
-
-    def scroll_down(self, from_line, scroll_height, back_attr):
-        """Scroll the screen down between from_line and scroll_height."""
-
-    def put_glyph(self, pagenum, row, col, char, is_fullwidth, fore, back, blink, underline):
-        """Put a character at a given position."""
-
-    def build_glyphs(self, new_dict):
-        """Build a dict of glyphs for use in text mode."""
-
-    def set_cursor_shape(self, width, height, from_line, to_line):
+    def set_cursor_shape(self, from_line, to_line):
         """Build a sprite for the cursor."""
 
-    def put_pixel(self, pagenum, x, y, index):
-        """Put a pixel on the screen; callback to empty character buffer."""
-
-    def fill_rect(self, pagenum, x0, y0, x1, y1, index):
-        """Fill a rectangle in a solid attribute."""
-
-    def fill_interval(self, pagenum, x0, x1, y, index):
-        """Fill a scanline interval in a solid attribute."""
-
-    def put_interval(self, pagenum, x, y, colours):
-        """Write a list of attributes to a scanline interval."""
-
-    def put_rect(self, pagenum, x0, y0, x1, y1, array):
-        """Apply numpy array [y][x] of attribytes to an area."""
+    def update(self, row, col, unicode_matrix, attr_matrix, y0, x0, sprite):
+        """Put text or pixels at a given position."""
