@@ -2,27 +2,24 @@
 PC-BASIC - video_pygame.py
 Graphical interface based on PyGame
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 import logging
+import ctypes
 
-try:
+if False:
+    # for detection by packagers
     import pygame
-except ImportError:
-    pygame = None
-
-try:
-    import numpy
-except ImportError:
-    numpy = None
 
+from ..compat import iteritems, unichr
 from ..basic.base import signals
 from ..basic.base import scancode
+from ..basic.base import bytematrix
 from ..basic.base.eascii import as_unicode as uea
-from ..data.resources import ICON
-from ..compat import WIN32, MACOS
+from ..compat import WIN32, MACOS, PY2
+from ..compat import set_dpi_aware
 from .video import VideoPlugin
 from .base import video_plugins, InitFailed, EnvironmentCache, NOKILL_MESSAGE
 from . import clipboard
@@ -36,6 +33,9 @@ BLINK_CYCLES = 5
 BLINK_TIME = 120
 CYCLE_TIME = BLINK_TIME // BLINK_CYCLES
 
+# blank icon
+BLANK_ICON = ((0,) * 16) * 16
+
 
 @video_plugins.register('pygame')
 class VideoPygame(VideoPlugin):
@@ -43,12 +43,17 @@ class VideoPygame(VideoPlugin):
 
     def __init__(
             self, input_queue, video_queue,
-            caption=u'', icon=ICON,
+            caption=u'', icon=BLANK_ICON,
             scaling=None, dimensions=None, aspect_ratio=(4, 3), border_width=0, fullscreen=False,
             prevent_close=False, mouse_clipboard=True,
             **kwargs
         ):
         """Initialise pygame interface."""
+        logging.warning('The `pygame` interface is deprecated, please use the `graphical` interface instead.')
+        try:
+            _import_pygame()
+        except ImportError:
+            raise InitFailed('Module `pygame` not found')
         VideoPlugin.__init__(self, input_queue, video_queue)
         # request smooth scaling
         self._smooth = scaling == 'smooth'
@@ -60,10 +65,8 @@ class VideoPygame(VideoPlugin):
         self.fullscreen = fullscreen
         self._has_window = False
         # set state objects to whatever is now in state (may have been unpickled)
-        if not pygame:
-            raise InitFailed('Module `pygame` not found')
-        if not numpy:
-            raise InitFailed('Module `numpy` not found')
+        # Windows 10 - set to DPI aware to avoid scaling twice on HiDPI screens
+        set_dpi_aware()
         # ensure we have the correct video driver for SDL 1.2
         # pygame sets this on import, but if we've tried SDL2 we've had to
         # reset this value
@@ -79,19 +82,18 @@ class VideoPygame(VideoPlugin):
             raise InitFailed('No suitable display driver: %s' % e)
         # display & border
         # display buffer
-        self.canvas = []
+        self.canvas = None
         # border attribute
         self.border_attr = 0
         # palette and colours
-        # composite colour artifacts
-        self._composite = False
         # working palette - attribute index in blue channel
         self.work_palette = [(0, 0, index) for index in range(256)]
         # display palettes for blink states 0, 1
         self._palette = [None, None]
-        self._saved_palette = [None, None]
+        # composite colour artifacts
+        self._pixel_packing = False
         # text attributes supported
-        self.mode_has_blink = True
+        self._palette_blinks = True
         # update cycle
         # update flag
         self.busy = False
@@ -106,36 +108,40 @@ class VideoPygame(VideoPlugin):
         self.last_col = 1
         # cursor is visible
         self.cursor_visible = True
+        self.cursor_attr = 7
         # buffer for text under cursor
         self.under_top_left = None
         # fonts
-        # prebuilt glyphs
-        self.glyph_dict = {}
         # joystick and mouse
         # available joysticks
         self.joysticks = []
         # get physical screen dimensions (needs to be called before set_mode)
         display_info = pygame.display.Info()
         self._window_sizer = window.WindowSizer(
-                display_info.current_w, display_info.current_h,
-                scaling, dimensions, aspect_ratio, border_width, fullscreen)
+            display_info.current_w, display_info.current_h,
+            scaling, dimensions, aspect_ratio, border_width,
+        )
         # determine initial display size
-        self.display_size = self._window_sizer.find_display_size(720, 400)
+        self._window_sizer.set_canvas_size(720, 400)
         self._set_icon(icon)
         try:
-            self._resize_display(*self.display_size)
+            self._resize_display()
         except pygame.error as e:
             self._close_pygame()
             raise InitFailed('Could not initialise display: %s' % e)
         if self._smooth and self.display.get_bitsize() < 24:
             logging.warning(
-                    'Smooth scaling not available on this display (depth %d < 24)',
-                    self.display.get_bitsize())
+                'Smooth scaling not available on this display (depth %d < 24)',
+                self.display.get_bitsize()
+            )
             self._smooth = False
-        pygame.display.set_caption(self.caption.encode('utf-8', 'replace'))
+        if PY2: # pragma: no cover
+            pygame.display.set_caption(self.caption.encode('utf-8', 'replace'))
+        else:
+            pygame.display.set_caption(self.caption)
         pygame.key.set_repeat(500, 24)
         # load an all-black 16-colour game palette to get started
-        self.set_palette([(0,0,0)]*16, None)
+        self.set_palette([((0,0,0), (0,0,0), False, False)]*16, None)
         pygame.joystick.init()
         self.joysticks = [pygame.joystick.Joystick(x) for x in range(pygame.joystick.get_count())]
         for j in self.joysticks:
@@ -146,8 +152,7 @@ class VideoPygame(VideoPlugin):
                 self._input_queue.put(signals.Event(signals.STICK_MOVED, (joy, axis, 128)))
         # mouse setups
         self._mouse_clip = mouse_clipboard
-        self.move_cursor(0, 0)
-        self.set_page(0, 0)
+        self.cursor_row, self.cursor_col = 1, 1
         # set_mode should be first event on queue
         self.f11_active = False
         self.clipboard_handler = get_clipboard_handler()
@@ -174,7 +179,9 @@ class VideoPygame(VideoPlugin):
         height, width = len(mask), len(mask[0])
         icon = pygame.Surface((width, height), depth=8) # pylint: disable=E1121,E1123
         icon.fill(0)
-        icon.blit(glyph_to_surface(mask), (0, 0, width, height))
+        array = bytematrix.ByteMatrix(height, width, mask)
+        #icon.blit(glyph_to_surface(mask), (0, 0, width, height))
+        pygame.surfarray.blit_array(icon, _BufferWrapper(array))
         icon.set_palette_at(0, (0, 0, 0))
         icon.set_palette_at(1, (0xff, 0xff, 0xff))
         pygame.transform.scale2x(icon)
@@ -200,70 +207,82 @@ class VideoPygame(VideoPlugin):
                     if event.button == 1:
                         # LEFT button: copy
                         pos = self._window_sizer.normalise_pos(*event.pos)
-                        self.clipboard.start(1 + pos[1] // self.font_height,
-                                1 + (pos[0]+self.font_width//2) // self.font_width)
+                        self.clipboard.start(
+                            1 + pos[1] // self.font_height,
+                            1 + (pos[0]+self.font_width//2) // self.font_width
+                        )
                     elif event.button == 2:
                         # MIDDLE button: paste
-                        text = self.clipboard_handler.paste(mouse=True)
+                        text = self.clipboard_handler.paste()
                         self.clipboard.paste(text)
+                    self.busy = True
                 if event.button == 1:
                     # right mouse button is a pen press
-                    self._input_queue.put(signals.Event(signals.PEN_DOWN,
-                                                self._window_sizer.normalise_pos(*event.pos)))
+                    self._input_queue.put(signals.Event(
+                        signals.PEN_DOWN, self._window_sizer.normalise_pos(*event.pos)
+                    ))
             elif event.type == pygame.MOUSEBUTTONUP:
                 self._input_queue.put(signals.Event(signals.PEN_UP))
                 if self._mouse_clip and event.button == 1:
-                    self.clipboard.copy(mouse=True)
+                    self.clipboard.copy()
                     self.clipboard.stop()
+                    self.busy = True
             elif event.type == pygame.MOUSEMOTION:
                 pos = self._window_sizer.normalise_pos(*event.pos)
                 self._input_queue.put(signals.Event(signals.PEN_MOVED, pos))
                 if self.clipboard.active():
-                    self.clipboard.move(1 + pos[1] // self.font_height,
-                           1 + (pos[0]+self.font_width//2) // self.font_width)
+                    self.clipboard.move(
+                        1 + pos[1] // self.font_height,
+                        1 + (pos[0]+self.font_width//2) // self.font_width
+                    )
+                    self.busy = True
             elif event.type == pygame.JOYBUTTONDOWN:
-                self._input_queue.put(signals.Event(signals.STICK_DOWN,
-                                                      (event.joy, event.button)))
+                self._input_queue.put(signals.Event(
+                    signals.STICK_DOWN, (event.joy, event.button)
+                ))
             elif event.type == pygame.JOYBUTTONUP:
-                self._input_queue.put(signals.Event(signals.STICK_UP,
-                                                      (event.joy, event.button)))
+                self._input_queue.put(signals.Event(
+                    signals.STICK_UP, (event.joy, event.button)
+                ))
             elif event.type == pygame.JOYAXISMOTION:
-                self._input_queue.put(signals.Event(signals.STICK_MOVED,
-                                                      (event.joy, event.axis,
-                                                      int(event.value*127 + 128))))
+                self._input_queue.put(signals.Event(
+                    signals.STICK_MOVED, (event.joy, event.axis, int(event.value*127 + 128))
+                ))
             elif event.type == pygame.VIDEORESIZE:
                 if not self.fullscreen:
-                    self._resize_display(event.w, event.h)
+                    self._window_sizer.set_display_size(event.w, event.h)
+                    self._resize_display()
             elif event.type == pygame.QUIT:
                 if self._nokill:
                     self.set_caption_message(NOKILL_MESSAGE)
                 else:
-                    self._input_queue.put(signals.Event(signals.KEYB_QUIT))
+                    self._input_queue.put(signals.Event(signals.QUIT))
 
     def _handle_key_down(self, e):
         """Handle key-down event."""
         # get scancode
-        scan = key_to_scan.get(e.key, None)
+        scan = KEY_TO_SCAN.get(e.key, None)
         # get modifiers
-        mod = [s for m, s in mod_to_scan.iteritems() if e.mod & m]
+        mod = [s for m, s in iteritems(MOD_TO_SCAN) if e.mod & m]
         # compensate for missing l-alt down events (needed at least on Ubuntu Unity)
         if (e.mod & pygame.KMOD_LALT) and not self._alt_key_down:
             # insert an alt keydown event before the alt-modified keydown
             fake_event = pygame.event.Event(
-                    pygame.KEYDOWN, scancode=0, key=pygame.K_LALT, unicode=u'', mod=0)
+                pygame.KEYDOWN, scancode=0, key=pygame.K_LALT, unicode=u'', mod=0
+            )
             self._handle_key_down(fake_event)
         if e.key == pygame.K_LALT:
             self._alt_key_down = True
         # get eascii
         try:
             if e.mod & pygame.KMOD_LALT:
-                c = alt_key_to_eascii[e.key]
+                c = ALT_KEY_TO_EASCII[e.key]
             elif e.mod & pygame.KMOD_CTRL:
-                c = ctrl_key_to_eascii[e.key]
+                c = CTRL_KEY_TO_EASCII[e.key]
             elif e.mod & pygame.KMOD_SHIFT:
-                c = shift_key_to_eascii[e.key]
+                c = SHIFT_KEY_TO_EASCII[e.key]
             else:
-                c = key_to_eascii[e.key]
+                c = KEY_TO_EASCII[e.key]
         except KeyError:
             key = e.key
             if e.mod & pygame.KMOD_CTRL and key >= ord('a') and key <= ord('z'):
@@ -280,7 +299,8 @@ class VideoPygame(VideoPlugin):
             # F11+f to toggle fullscreen mode
             if e.key == pygame.K_f:
                 self.fullscreen = not self.fullscreen
-                self._resize_display(*self._window_sizer.find_display_size(*self.size))
+                self._window_sizer.set_canvas_size(*self.size, fullscreen=self.fullscreen)
+                self._resize_display()
             self.clipboard.handle_key(scan, c)
             self.busy = True
         else:
@@ -290,12 +310,11 @@ class VideoPygame(VideoPlugin):
             # fix missing ascii for numeric keypad on Windows
             if e.mod & pygame.KMOD_NUM:
                 try:
-                    c = key_to_eascii_num[e.key]
+                    c = KEY_TO_EASCII_NUM[e.key]
                 except KeyError:
                     pass
             # insert into keyboard queue
-            self._input_queue.put(signals.Event(
-                                    signals.KEYB_DOWN, (c, scan, mod)))
+            self._input_queue.put(signals.Event(signals.KEYB_DOWN, (c, scan, mod)))
 
     def _handle_key_up(self, e):
         """Handle key-up event."""
@@ -303,7 +322,8 @@ class VideoPygame(VideoPlugin):
         if (e.key == pygame.K_LALT or e.mod & pygame.KMOD_LALT) and not self._alt_key_down:
             # insert an alt keydown event before the alt-modified keydown
             fake_event = pygame.event.Event(
-                    pygame.KEYDOWN, scancode=0, key=pygame.K_LALT, unicode=u'', mod=0)
+                pygame.KEYDOWN, scancode=0, key=pygame.K_LALT, unicode=u'', mod=0
+            )
             self._handle_key_down(fake_event)
             if e.key != pygame.K_LALT:
                 # insert a keydown before the keyup
@@ -318,8 +338,7 @@ class VideoPygame(VideoPlugin):
             self.f11_active = False
         # last key released gets remembered
         try:
-            self._input_queue.put(signals.Event(
-                                    signals.KEYB_UP, (key_to_scan[e.key],)))
+            self._input_queue.put(signals.Event(signals.KEYB_UP, (KEY_TO_SCAN[e.key],)))
         except KeyError:
             pass
 
@@ -336,12 +355,13 @@ class VideoPygame(VideoPlugin):
         if not self._has_window:
             return
         self.blink_state = 0
-        if self.mode_has_blink:
+        if self._palette_blinks or self.text_cursor:
             self.blink_state = 0 if self._cycle < BLINK_CYCLES * 2 else 1
             if self._cycle % BLINK_CYCLES == 0:
                 self.busy = True
         if self.cursor_visible and (
-                (self.cursor_row != self.last_row) or (self.cursor_col != self.last_col)):
+                (self.cursor_row != self.last_row) or (self.cursor_col != self.last_col)
+            ):
             self.busy = True
         tock = pygame.time.get_ticks()
         if tock - self.last_cycle >= CYCLE_TIME:
@@ -356,58 +376,74 @@ class VideoPygame(VideoPlugin):
     def _do_flip(self):
         """Draw the canvas to the screen."""
         # create the screen that will be stretched onto the display
-        border_x, border_y = self._window_sizer.border_start()
+        border_x, border_y = self._window_sizer.border_shift
         # surface depth and flags match those of canvas
         # pylint: disable=E1121,E1123
         screen = pygame.Surface(
             (self.size[0] + 2*border_x, self.size[1] + 2*border_y),
-            0, self.canvas[self.vpagenum])
+            0, self.canvas
+        )
         screen.set_palette(self.work_palette)
         # border colour
         border_colour = pygame.Color(0, 0, self.border_attr % self.num_fore_attrs)
         screen.fill(border_colour)
-        screen.blit(self.canvas[self.vpagenum], (border_x, border_y))
+        screen.blit(self.canvas, (border_x, border_y))
         # subsurface referencing the canvas area
         workscreen = screen.subsurface((border_x, border_y, self.size[0], self.size[1]))
         self._draw_cursor(workscreen)
         if self.clipboard.active():
             create_feedback(workscreen, self.clipboard.selection_rect)
-        if self._composite:
-            screen = apply_composite_artifacts(screen, 4//self.bitsperpixel)
+        if self._pixel_packing:
+            screen = apply_composite_artifacts(screen, *self._pixel_packing)
         screen.set_palette(self._palette[self.blink_state])
+        letterbox = pygame.Rect(
+            self._window_sizer.letterbox_shift, self._window_sizer.window_size
+        )
         if self._smooth:
-            pygame.transform.smoothscale(screen.convert(self.display),
-                                         self.display.get_size(), self.display)
+            # smoothscale to subsurface does not work correctly, so create a new surface and blit
+            dest_surf = pygame.Surface(self._window_sizer.window_size) #, self.display)
+            pygame.transform.smoothscale(
+                screen.convert(dest_surf), self._window_sizer.window_size, dest_surf
+            )
+            self.display.blit(dest_surf, letterbox)
         else:
-            pygame.transform.scale(screen.convert(self.display),
-                                   self.display.get_size(), self.display)
+            dest_surf = self.display.subsurface(letterbox)
+            pygame.transform.scale(
+                screen.convert(dest_surf), self._window_sizer.window_size, dest_surf
+            )
         pygame.display.flip()
 
     def _draw_cursor(self, screen):
         """Draw the cursor on the surface provided."""
-        if not self.cursor_visible or self.vpagenum != self.apagenum:
+        if not self.cursor_visible:
             return
         # copy screen under cursor
         self.under_top_left = (
-                (self.cursor_col-1) * self.font_width, (self.cursor_row-1) * self.font_height)
+            (self.cursor_col-1) * self.font_width,
+            (self.cursor_row-1) * self.font_height
+        )
         under_char_area = pygame.Rect(
-                (self.cursor_col-1) * self.font_width, (self.cursor_row-1) * self.font_height,
-                self.cursor_width, self.font_height)
-        if self.text_mode:
+            (self.cursor_col-1) * self.font_width, (self.cursor_row-1) * self.font_height,
+            self.cursor_width, self.font_height
+        )
+        if self.text_cursor:
             # cursor is visible - to be done every cycle between 5 and 10, 15 and 20
             if self._cycle // BLINK_CYCLES in (1, 3):
-                screen.blit(self.cursor, (
-                        (self.cursor_col-1) * self.font_width,
-                        (self.cursor_row-1) * self.font_height) )
+                screen.blit(
+                    self.cursor,
+                    ((self.cursor_col-1) * self.font_width, (self.cursor_row-1) * self.font_height)
+                )
         else:
             index = self.cursor_attr % self.num_fore_attrs
             # reference the destination area
             dest_array = pygame.surfarray.pixels2d(
-                    screen.subsurface(pygame.Rect(
-                        (self.cursor_col-1) * self.font_width,
-                        (self.cursor_row-1) * self.font_height + self.cursor_from,
-                        self.cursor_width,
-                        self.cursor_to - self.cursor_from + 1)))
+                screen.subsurface(pygame.Rect(
+                    (self.cursor_col-1) * self.font_width,
+                    (self.cursor_row-1) * self.font_height + self.cursor_from,
+                    self.cursor_width,
+                    self.cursor_to - self.cursor_from + 1
+                ))
+            )
             dest_array ^= index
         self.last_row = self.cursor_row
         self.last_col = self.cursor_col
@@ -415,14 +451,13 @@ class VideoPygame(VideoPlugin):
     ###########################################################################
     # miscellaneous helper functions
 
-    def _resize_display(self, width, height):
+    def _resize_display(self):
         """Change the display size."""
         if self.fullscreen:
-            flags = pygame.FULLSCREEN | pygame.NOFRAME
+            flags = pygame.NOFRAME
         else:
             flags = pygame.RESIZABLE
-        self.display = pygame.display.set_mode((width, height), flags)
-        self._window_sizer.window_size = width, height
+        self.display = pygame.display.set_mode(self._window_sizer.display_size, flags)
         # load display if requested
         self.busy = True
 
@@ -430,32 +465,26 @@ class VideoPygame(VideoPlugin):
     ###########################################################################
     # signal handlers
 
-    def set_mode(self, mode_info):
+    def set_mode(self, canvas_height, canvas_width, text_height, text_width):
         """Initialise a given text or graphics mode."""
-        self.text_mode = mode_info.is_text_mode
-        # unpack mode info struct
-        self.font_height = mode_info.font_height
-        self.font_width = mode_info.font_width
-        self.num_pages = mode_info.num_pages
-        self.mode_has_blink = mode_info.has_blink
-        if not self.text_mode:
-            self.bitsperpixel = mode_info.bitsperpixel
+        # set display geometry
+        self.font_height = -(-canvas_height // text_height)
+        self.font_width = canvas_width // text_width
         # logical size
-        self.size = (mode_info.pixel_width, mode_info.pixel_height)
-        self._window_sizer.size = self.size
-        self._resize_display(*self._window_sizer.find_display_size(*self.size))
+        self.size = canvas_width, canvas_height
+        self._window_sizer.set_canvas_size(*self.size, fullscreen=self.fullscreen)
+        self._resize_display()
         # set standard cursor
-        self.set_cursor_shape(self.font_width, self.font_height,
-                              0, self.font_height)
+        self.cursor_width = self.font_width
+        self.set_cursor_shape(0, self.font_height)
         # whole screen (blink on & off)
-        self.canvas = [ pygame.Surface(self.size, depth=8) # pylint: disable=E1121,E1123
-                        for _ in range(self.num_pages)]
-        for i in range(self.num_pages):
-            self.canvas[i].set_palette(self.work_palette)
+        self.canvas = pygame.Surface(self.size, depth=8) # pylint: disable=E1121,E1123
+        self.canvas.set_palette(self.work_palette)
         # initialise clipboard
         self.clipboard = clipboard.ClipboardInterface(
-                self.clipboard_handler, self._input_queue,
-                mode_info.width, mode_info.height, self.font_width, self.font_height, self.size)
+            self.clipboard_handler, self._input_queue,
+            text_width, text_height, self.font_width, self.font_height, self.size
+        )
         self.busy = True
         self._has_window = True
 
@@ -464,23 +493,21 @@ class VideoPygame(VideoPlugin):
         title = self.caption + (u' - ' + msg if msg else u'')
         pygame.display.set_caption(title.encode('utf-8', 'replace'))
 
-    def set_clipboard_text(self, text, mouse):
+    def set_clipboard_text(self, text):
         """Put text on the clipboard."""
-        self.clipboard_handler.copy(text, mouse)
+        self.clipboard_handler.copy(text)
 
-    def set_palette(self, rgb_palette_0, rgb_palette_1):
+    def set_palette(self, attributes, pack_pixels):
         """Build the palette."""
-        self.num_fore_attrs = min(16, len(rgb_palette_0))
-        self.num_back_attrs = min(8, self.num_fore_attrs)
-        rgb_palette_1 = rgb_palette_1 or rgb_palette_0
+        self.num_fore_attrs = 16
+        self.num_back_attrs = 8
         # fill up the 8-bit palette with all combinations we need
         # blink states: 0 light up, 1 light down
         # bottom 128 are non-blink, top 128 blink to background
-        self._palette[0] = rgb_palette_0[:self.num_fore_attrs] * (256//self.num_fore_attrs)
-        self._palette[1] = rgb_palette_1[:self.num_fore_attrs] * (128//self.num_fore_attrs)
-        for b in rgb_palette_1[:self.num_back_attrs] * (
-                    128 // self.num_fore_attrs // self.num_back_attrs):
-            self._palette[1] += [b]*self.num_fore_attrs
+        self._palette[0] = [_fore for _fore, _, _, _ in attributes]
+        self._palette[1] = [_back if _blink else _fore for _fore, _back, _blink, _ in attributes]
+        self._palette_blinks = self._palette[0] != self._palette[1]
+        self._pixel_packing = pack_pixels
         self.busy = True
 
     def set_border_attr(self, attr):
@@ -488,159 +515,90 @@ class VideoPygame(VideoPlugin):
         self.border_attr = attr
         self.busy = True
 
-    def set_composite(self, on, composite_colors):
-        """Enable/disable composite artifacts."""
-        if on != self._composite:
-            self._palette, self._saved_palette = self._saved_palette, self._palette
-        if on:
-            self._palette = [composite_colors] * 2
-        self._composite = on
-        self.busy = True
-
     def clear_rows(self, back_attr, start, stop):
         """Clear a range of screen rows."""
         bg = (0, 0, back_attr)
         scroll_area = pygame.Rect(
-                0, (start-1)*self.font_height, self.size[0], (stop-start+1)*self.font_height)
-        self.canvas[self.apagenum].fill(bg, scroll_area)
+            0, (start-1)*self.font_height, self.size[0], (stop-start+1)*self.font_height
+        )
+        self.canvas.fill(bg, scroll_area)
         self.busy = True
 
-    def set_page(self, vpage, apage):
-        """Set the visible and active page."""
-        self.vpagenum, self.apagenum = vpage, apage
-        self.busy = True
-
-    def copy_page(self, src, dst):
-        """Copy source to destination page."""
-        self.canvas[dst].blit(self.canvas[src], (0, 0))
-        self.busy = True
-
-    def show_cursor(self, cursor_on):
+    def show_cursor(self, cursor_on, cursor_blinks):
         """Change visibility of cursor."""
         self.cursor_visible = cursor_on
+        self.text_cursor = cursor_blinks
         self.busy = True
 
-    def move_cursor(self, crow, ccol):
+    def move_cursor(self, row, col, attr, width):
         """Move the cursor to a new position."""
-        self.cursor_row, self.cursor_col = crow, ccol
-
-    def set_cursor_attr(self, attr):
-        """Change attribute of cursor."""
+        self.cursor_row, self.cursor_col = row, col
+        # set attribute
         self.cursor_attr = attr % self.num_fore_attrs
-        self.cursor.set_palette_at(254, pygame.Color(0, self.cursor_attr, self.cursor_attr))
+        # set width
+        if width != self.cursor_width:
+            self.cursor_width = width
+            self._rebuild_cursor()
+        else:
+            self.cursor.set_palette_at(254, pygame.Color(0, self.cursor_attr, self.cursor_attr))
 
-    def scroll_up(self, from_line, scroll_height, back_attr):
-        """Scroll the screen up between from_line and scroll_height."""
+    def scroll(self, direction, from_line, scroll_height, back_attr):
+        """Scroll the screen between from_line and scroll_height."""
         temp_scroll_area = pygame.Rect(
             0, (from_line-1)*self.font_height,
             self.size[0], (scroll_height-from_line+1) * self.font_height
         )
         # scroll
-        self.canvas[self.apagenum].set_clip(temp_scroll_area)
-        self.canvas[self.apagenum].scroll(0, -self.font_height)
+        self.canvas.set_clip(temp_scroll_area)
+        self.canvas.scroll(0, direction * self.font_height)
         # empty new line
         bg = (0, 0, back_attr)
-        self.canvas[self.apagenum].fill(
+        self.canvas.fill(
             bg, (0, (scroll_height-1) * self.font_height, self.size[0], self.font_height)
         )
-        self.canvas[self.apagenum].set_clip(None)
+        self.canvas.set_clip(None)
         self.busy = True
 
-    def scroll_down(self, from_line, scroll_height, back_attr):
-        """Scroll the screen down between from_line and scroll_height."""
-        temp_scroll_area = pygame.Rect(
-            0, (from_line-1) * self.font_height,
-            self.size[0], (scroll_height-from_line+1) * self.font_height
-        )
-        self.canvas[self.apagenum].set_clip(temp_scroll_area)
-        self.canvas[self.apagenum].scroll(0, self.font_height)
-        # empty new line
-        bg = (0, 0, back_attr)
-        self.canvas[self.apagenum].fill(
-            bg, (0, (from_line-1) * self.font_height, self.size[0], self.font_height)
-        )
-        self.canvas[self.apagenum].set_clip(None)
-        self.busy = True
-
-    def put_glyph(self, pagenum, row, col, cp, is_fullwidth, fore, back, blink, underline):
-        """Put a single-byte character at a given position."""
-        if not self.text_mode:
-            # in graphics mode, a put_rect call does the actual drawing
-            return
-        color = (0, 0, fore + self.num_fore_attrs*back + 128*blink)
-        bg = (0, 0, back)
-        x0, y0 = (col-1)*self.font_width, (row-1)*self.font_height
-        if cp == u'\0':
-            # guaranteed to be blank, saves time on some BLOADs
-            self.canvas[pagenum].fill(bg, (x0, y0, self.font_width, self.font_height))
-        else:
-            try:
-                glyph = self.glyph_dict[cp]
-            except KeyError:
-                if u'\0' not in self.glyph_dict:
-                    logging.error('No glyph received for code point 0')
-                    return
-                logging.warning('No glyph received for code point %s', hex(ord(cp)))
-                glyph = self.glyph_dict[u'\0']
-            if glyph.get_palette_at(0) != bg:
-                glyph.set_palette_at(0, bg)
-            if glyph.get_palette_at(1) != color:
-                glyph.set_palette_at(1, color)
-            self.canvas[pagenum].blit(glyph, (x0, y0))
-        if underline:
-            self.canvas[pagenum].fill(color, (x0, y0 + self.font_height - 1, self.font_width, 1))
-        self.busy = True
-
-    def build_glyphs(self, new_dict):
-        """Build a dict of glyphs for use in text mode."""
-        for char, glyph in new_dict.iteritems():
-            self.glyph_dict[char] = glyph_to_surface(glyph)
-
-    def set_cursor_shape(self, width, height, from_line, to_line):
+    def set_cursor_shape(self, from_line, to_line):
         """Build a sprite for the cursor."""
-        self.cursor_width = width
         self.cursor_from, self.cursor_to = from_line, to_line
+        self._rebuild_cursor()
+
+    def _rebuild_cursor(self):
+        """Rebuild cursor surface."""
+        height = self.font_height
+        width = self.cursor_width
+        from_line, to_line = self.cursor_from, self.cursor_to
         self.cursor = pygame.Surface((width, height), depth=8) # pylint: disable=E1121,E1123
         color, bg = 254, 255
         self.cursor.set_colorkey(bg)
         self.cursor.fill(bg)
         self.cursor.fill(color, (0, from_line, width, min(to_line-from_line+1, height-from_line)))
+        self.cursor.set_palette_at(254, pygame.Color(0, self.cursor_attr, self.cursor_attr))
         self.busy = True
 
-    def put_pixel(self, pagenum, x, y, index):
-        """Put a pixel on the screen; callback to empty character buffer."""
-        self.canvas[pagenum].set_at((x,y), index)
-        self.busy = True
-
-    def fill_rect(self, pagenum, x0, y0, x1, y1, index):
-        """Fill a rectangle in a solid attribute."""
-        rect = pygame.Rect(x0, y0, x1-x0+1, y1-y0+1)
-        self.canvas[pagenum].fill(index, rect)
-        self.busy = True
-
-    def fill_interval(self, pagenum, x0, x1, y, index):
-        """Fill a scanline interval in a solid attribute."""
-        dx = x1 - x0 + 1
-        self.canvas[pagenum].fill(index, (x0, y, dx, 1))
-        self.busy = True
-
-    def put_interval(self, pagenum, x, y, colours):
-        """Write a list of attributes to a scanline interval."""
-        # reference the interval on the canvas
-        pygame.surfarray.pixels2d(self.canvas[pagenum]
-                )[x:x+len(colours), y] = numpy.array(colours).astype(int)
-        self.busy = True
-
-    def put_rect(self, pagenum, x0, y0, x1, y1, array):
-        """Apply numpy array [y][x] of attribytes to an area."""
-        if (x1 < x0) or (y1 < y0):
-            return
+    def update(self, row, col, unicode_matrix, attr_matrix, y0, x0, array):
+        """Put text or pixels at a given position."""
+        if y0 + array.height > self.size[1] or x0 +array. width > self.size[0]:
+            array = array[:self.size[1]-y0, :self.size[0]-x0]
         # reference the destination area
-        pygame.surfarray.pixels2d(self.canvas[pagenum].subsurface(
-            pygame.Rect(x0, y0, x1-x0+1, y1-y0+1)))[:] = numpy.array(array).T
+        subsurface = self.canvas.subsurface(pygame.Rect(x0, y0, array.width, array.height))
+        pygame.surfarray.blit_array(subsurface, _BufferWrapper(array))
         self.busy = True
 
 
+class _BufferWrapper(object):
+
+    def __init__(self, array):
+        self._buffer = ctypes.create_string_buffer(array.to_bytes())
+        self.__array_interface__ = dict(
+            shape=(array.width, array.height),
+            strides=(1, array.width),
+            typestr='|u1', version=3, data=(ctypes.addressof(self._buffer), False)
+        )
+
+
+
 ###############################################################################
 # clipboard handling
 
@@ -663,12 +621,9 @@ class PygameClipboard(clipboard.Clipboard):
                 logging.warning('PyGame.Scrap clipboard handling module not found.')
             self.ok = False
 
-    def copy(self, text, mouse=False):
+    def copy(self, text):
         """Put unicode text on clipboard."""
-        if mouse:
-            pygame.scrap.set_mode(pygame.SCRAP_SELECTION)
-        else:
-            pygame.scrap.set_mode(pygame.SCRAP_CLIPBOARD)
+        pygame.scrap.set_mode(pygame.SCRAP_CLIPBOARD)
         try:
             if WIN32:
                 # on Windows, encode as utf-16 without FF FE byte order mark and null-terminate
@@ -678,16 +633,14 @@ class PygameClipboard(clipboard.Clipboard):
                 )
             else:
                 pygame.scrap.put(pygame.SCRAP_TEXT, text.encode('utf-8', 'replace'))
-        except KeyError:
-            logging.debug('Clipboard copy failed for clip %s', repr(text))
+        except Exception as e:# KeyError:
+            logging.debug('Clipboard copy failed for clip %r: %s', text, e)
 
-    def paste(self, mouse=False):
+    def paste(self):
         """Return unicode text from clipboard."""
-        if mouse:
-            pygame.scrap.set_mode(pygame.SCRAP_SELECTION)
-        else:
-            pygame.scrap.set_mode(pygame.SCRAP_CLIPBOARD)
+        pygame.scrap.set_mode(pygame.SCRAP_CLIPBOARD)
         us = u''
+        s = b''
         available = pygame.scrap.get_types()
         for text_type in self.text:
             if text_type not in available:
@@ -713,11 +666,13 @@ class PygameClipboard(clipboard.Clipboard):
 
 def get_clipboard_handler():
     """Get a working Clipboard handler object."""
-    # Pygame.Scrap doesn't work on OSX
+    # Pygame.Scrap doesn't work on OSX and crashes on Linux
     if MACOS:
         handler = clipboard.MacClipboard()
-    else:
+    elif WIN32:
         handler = PygameClipboard()
+    else:
+        handler = clipboard.XClipboard()
     if not handler.ok:
         logging.warning('Clipboard copy and paste not available.')
         handler = clipboard.Clipboard()
@@ -734,11 +689,21 @@ def create_feedback(surface, selection_rects):
 
 
 ###############################################################################
+# import PyGame and define constants
+
+pygame = None
 
 
-if pygame:
+def _import_pygame():
+    """Import pygame and define constants."""
+    global pygame
+    global KEY_TO_SCAN, MOD_TO_SCAN, KEY_TO_EASCII_NUM
+    global KEY_TO_EASCII, SHIFT_KEY_TO_EASCII, CTRL_KEY_TO_EASCII, ALT_KEY_TO_EASCII
+
+    import pygame
+
     # these are PC keyboard scancodes
-    key_to_scan = {
+    KEY_TO_SCAN = {
         # top row
         pygame.K_ESCAPE: scancode.ESCAPE, pygame.K_1: scancode.N1,
         pygame.K_2: scancode.N2, pygame.K_3: scancode.N3,
@@ -806,8 +771,7 @@ if pygame:
         pygame.K_BREAK: scancode.BREAK,
     }
 
-
-    key_to_eascii = {
+    KEY_TO_EASCII = {
         pygame.K_F1: uea.F1,
         pygame.K_F2: uea.F2,
         pygame.K_F3: uea.F3,
@@ -838,7 +802,7 @@ if pygame:
         pygame.K_DELETE: uea.DELETE,
     }
 
-    shift_key_to_eascii = {
+    SHIFT_KEY_TO_EASCII = {
         pygame.K_F1: uea.SHIFT_F1,
         pygame.K_F2: uea.SHIFT_F2,
         pygame.K_F3: uea.SHIFT_F3,
@@ -870,7 +834,7 @@ if pygame:
         pygame.K_KP5: uea.SHIFT_KP5,
     }
 
-    ctrl_key_to_eascii = {
+    CTRL_KEY_TO_EASCII = {
         pygame.K_F1: uea.CTRL_F1,
         pygame.K_F2: uea.CTRL_F2,
         pygame.K_F3: uea.CTRL_F3,
@@ -901,7 +865,7 @@ if pygame:
         pygame.K_MINUS: uea.CTRL_MINUS,
     }
 
-    alt_key_to_eascii = {
+    ALT_KEY_TO_EASCII = {
         pygame.K_1: uea.ALT_1,
         pygame.K_2: uea.ALT_2,
         pygame.K_3: uea.ALT_3,
@@ -961,7 +925,7 @@ if pygame:
         pygame.K_KP5: uea.ALT_KP5,
     }
 
-    mod_to_scan = {
+    MOD_TO_SCAN = {
         pygame.KMOD_LSHIFT: scancode.LSHIFT,
         pygame.KMOD_RSHIFT: scancode.RSHIFT,
         pygame.KMOD_LCTRL: scancode.CTRL,
@@ -971,7 +935,7 @@ if pygame:
         #pygame.KMOD_RALT: scancode.ALT,
     }
 
-    key_to_eascii_num = {
+    KEY_TO_EASCII_NUM = {
         pygame.K_KP0: u'0',
         pygame.K_KP1: u'1',
         pygame.K_KP2: u'2',
@@ -985,16 +949,7 @@ if pygame:
     }
 
 
-def apply_composite_artifacts(screen, pixels=4):
+def apply_composite_artifacts(screen, bpp_out, bpp_in):
     """Process the canvas to apply composite colour artifacts."""
-    src_array = pygame.surfarray.array2d(screen)
-    return pygame.surfarray.make_surface(
-                    window.apply_composite_artifacts(src_array, pixels))
-
-
-def glyph_to_surface(glyph):
-    """Build a sprite surface for the given character glyph."""
-    glyph = numpy.asarray(glyph).T
-    surf = pygame.Surface(glyph.shape, depth=8) # pylint: disable=E1121,E1123
-    pygame.surfarray.pixels2d(surf)[:] = glyph
-    return surf
+    # not implemented
+    return screen
