@@ -2,33 +2,31 @@
 PC-BASIC - cassette.py
 Cassette Tape Device
 
-(c) 2015--2018 Rob Hagemans
+(c) 2015--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 import os
+import io
 import math
 import struct
 import logging
-import string
 from chunk import Chunk
-import io
+
+from ...compat import int2byte, iterchar, zip
 
 from ..base import error
 from ..base import tokens as tk
 from .devicebase import RawFile, TextFileBase, InputMixin, DeviceSettings, parse_protocol_string
 
 
-# mark bytes conversion explicitly
-int2byte = chr
-
 # file types (data, bsaved memory, protected, ascii, tokenised)
 TOKEN_TO_TYPE = {
     0: b'D', 1: b'M', 0xa0: b'P',
     0x20: b'P', 0x40: b'A', 0x80: b'B'
 }
 
-TYPE_TO_TOKEN = dict(reversed(item) for item in TOKEN_TO_TYPE.iteritems())
+TYPE_TO_TOKEN = dict(reversed(item) for item in TOKEN_TO_TYPE.items())
 
 
 #################################################################################
@@ -49,9 +47,9 @@ class CASDevice(object):
     """Cassette tape device (CASn:) """
 
     # control characters not allowed in file name on tape
-    _illegal_chars = set(map(int2byte, range(0x20)))
+    _illegal_chars = set(int2byte(_i) for _i in range(0x20))
 
-    def __init__(self, arg, screen):
+    def __init__(self, arg, console):
         """Initialise tape device."""
         addr, val = parse_protocol_string(arg)
         ext = val.split(u'.')[-1].upper()
@@ -60,7 +58,7 @@ class CASDevice(object):
         # by default, show messages
         self.is_quiet = False
         # console for messages
-        self.screen = screen
+        self.console = console
         try:
             if not val:
                 self.tapestream = None
@@ -89,7 +87,7 @@ class CASDevice(object):
             raise error.BASICError(error.DEVICE_UNAVAILABLE)
         if self.tapestream.is_open:
             raise error.BASICError(error.FILE_ALREADY_OPEN)
-        if set(param) & self._illegal_chars:
+        if set(iterchar(param)) & self._illegal_chars:
             # Cassette BASIC throws bad file NUMBER, for some reason.
             raise error.BASICError(error.BAD_FILE_NUMBER)
         try:
@@ -119,13 +117,13 @@ class CASDevice(object):
                     ):
                     message = b'%s.%s Found.' % (trunk, filetype)
                     if not self.is_quiet:
-                        self.screen.write_line(message)
+                        self.console.write_line(message)
                     logging.debug(timestamp(self.tapestream.counter()) + message)
                     return trunk, filetype, seg, offset, length
                 else:
                     message = b'%s.%s Skipped.' % (trunk, filetype)
                     if not self.is_quiet:
-                        self.screen.write_line(message)
+                        self.console.write_line(message)
                     logging.debug(timestamp(self.tapestream.counter()) + message)
         except EndOfTape:
             # reached end-of-tape without finding appropriate file
@@ -242,7 +240,7 @@ class CassetteStream(object):
         self.rwmode = 'r'
         while True:
             record = self._read_record(None)
-            if record and record[0] == b'\xa5':
+            if record and record[0:1] == b'\xa5':
                 break
             else:
                 # unknown record type
@@ -289,7 +287,7 @@ class CassetteStream(object):
         record = b''
         block_num = 0
         byte_count = 0
-        while byte_count < reclen or reclen is None:
+        while reclen is None or byte_count < reclen:
             data = self._read_block()
             record += data
             byte_count += len(data)
@@ -334,12 +332,12 @@ class CassetteStream(object):
     def _write_block(self, data):
         """Write a 256-byte block to tape."""
         # fill out short blocks with last byte
-        data += data[-1]*(256-len(data))
-        for b in data:
+        data += data[-1:]*(256-len(data))
+        for b in iterchar(data):
             self.bitstream.write_byte(ord(b))
         crc_word = crc(data)
         # crc is written big-endian
-        lo, hi = map(ord, struct.pack('<H', crc_word))
+        lo, hi = (ord(_b) for _b in iterchar(struct.pack('<H', crc_word)))
         self.bitstream.write_byte(hi)
         self.bitstream.write_byte(lo)
 
@@ -357,7 +355,7 @@ class CassetteStream(object):
             # 256 bytes less 1 length byte. CRC trailer comes after 256-byte block
             self.record_stream = io.BytesIO()
             record = self._read_record(256)
-            num_bytes = ord(record[0])
+            num_bytes = ord(record[0:1])
             record = record[1:]
             if num_bytes != 0:
                 record = record[:num_bytes-1]
@@ -430,8 +428,9 @@ class TapeBitStream(object):
     def read_intro(self):
         """Try to read intro; ensure image not empty."""
         for b in bytearray(self.intro):
-            c = self.read_byte()
-            if c == b'':
+            try:
+                c = self.read_byte()
+            except EndOfTape:
                 # empty or short file
                 return False
             if c != b:
@@ -485,7 +484,7 @@ class TapeBitStream(object):
         """Read a byte from the tape."""
         # NOTE: skip_start is ignored
         byte = 0
-        for i in xrange(8):
+        for i in range(8):
             bit = self.read_bit()
             if bit is None:
                 raise PulseError()
@@ -555,11 +554,11 @@ class CASBitStream(TapeBitStream):
             self.operating_mode = 'r'
             self.mask = 0x100
             try:
-                self.cas = open(self.cas_name, 'r+b')
+                self.cas = io.open(self.cas_name, 'r+b')
             except EnvironmentError:
-                self.cas = open(self.cas_name, 'rb')
+                self.cas = io.open(self.cas_name, 'rb')
             self.current_byte = self.cas.read(1)
-            if self.current_byte == '' or not self.read_intro():
+            if self.current_byte == b'' or not self.read_intro():
                 self.cas.close()
                 self._create()
         self.switch_mode(mode)
@@ -599,7 +598,7 @@ class CASBitStream(TapeBitStream):
         if self.mask <= 0:
             self.current_byte = self.cas.read(1)
             if not self.current_byte:
-                raise EndOfTape
+                raise EndOfTape()
             self.mask = 0x80
         if (ord(self.current_byte) & self.mask == 0):
             return 0
@@ -652,11 +651,11 @@ class CASBitStream(TapeBitStream):
         """Create a new CAS-file."""
         self.current_byte = b'\0'
         self.mask = 0x100
-        with open(self.cas_name, 'wb') as self.cas:
+        with io.open(self.cas_name, 'wb') as self.cas:
             self.operating_mode = 'w'
             self.current_byte = b'\0'
             self.write_intro()
-        self.cas = open(self.cas_name, 'r+b')
+        self.cas = io.open(self.cas_name, 'r+b')
         self.cas.seek(0, 2)
 
 
@@ -737,15 +736,15 @@ class WAVBitStream(TapeBitStream):
             self.framerate = 22050
             self.sampwidth = 1
             self.nchannels = 1
-            self.wav = open(self.filename, 'wb')
+            self.wav = io.open(self.filename, 'wb')
             self._write_wav_header()
             self.operating_mode = 'w'
         else:
             # open file for reading and find wave parameters
             try:
-                self.wav = open(self.filename, 'r+b')
+                self.wav = io.open(self.filename, 'r+b')
             except EnvironmentError:
-                self.wav = open(self.filename, 'rb')
+                self.wav = io.open(self.filename, 'rb')
             if not self._read_wav_header():
                 raise EndOfTape()
             self.operating_mode = 'r'
@@ -811,7 +810,7 @@ class WAVBitStream(TapeBitStream):
             length_up, length_dn = next(self.read_half), next(self.read_half)
         except StopIteration:
             self.read_half = self._gen_read_halfpulse()
-            raise EndOfTape
+            raise EndOfTape()
         if (length_up > self.halflength_max or length_dn > self.halflength_max or
                 length_up < self.halflength_min or length_dn < self.halflength_min):
             return None
@@ -840,10 +839,13 @@ class WAVBitStream(TapeBitStream):
             raise EndOfTape
         # convert MSBs to int (data stored little endian)
         # note that we simply throw away all the less significant bytes
-        frames = map(ord, frames[self.sampwidth-1::self.sampwidth])
+        frames = (ord(_c) for _c in iterchar(frames[self.sampwidth-1::self.sampwidth]))
         # sum frames over channels
         frames = map(sum, zip(*[iter(frames)]*self.nchannels))
-        frames = [ x-self.subtractor if x >= self.sub_threshold else x for x in frames ]
+        frames = [
+            _x-self.subtractor if _x >= self.sub_threshold else _x
+            for _x in frames
+        ]
         return self.filter.send(frames)
 
     def _gen_read_halfpulse(self):
@@ -872,7 +874,7 @@ class WAVBitStream(TapeBitStream):
 
     def write_pause(self, milliseconds):
         """Write a pause of given length to the tape."""
-        length = (milliseconds * self.framerate / 1000)
+        length = int(milliseconds * self.framerate / 1000)
         zero = {1: b'\x7f', 2: b'\x00\x00'}
         self.wav.write(zero[self.sampwidth] * self.nchannels * length)
         self.wav_pos += length
