@@ -2,7 +2,7 @@
 PC-BASIC - devices.diskfiles
 Disk Files
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -219,7 +219,7 @@ class FieldFile(TextFile):
 
     def get_buffer(self):
         """Get a copy of the contents of the buffer."""
-        return bytearray(self._field.view_buffer())
+        return bytearray(self._field.view_buffer()[:self._reclen])
 
     def write(self, bytestr, can_break=True):
         """Write bytes to buffer."""
