@@ -2,7 +2,7 @@
 PC-BASIC - formatter.py
 Formatted output handling
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -15,13 +15,13 @@ from .. import values
 class Formatter(object):
     """Output string formatter."""
 
-    def __init__(self, output, screen=None):
+    def __init__(self, output, console=None):
         """Initialise."""
-        self._screen = screen
+        self._console = console
         self._output = output
 
     def format(self, args):
-        """PRINT: Write expressions to screen or file."""
+        """PRINT: Write expressions to console or file."""
         newline = True
         for sep, value in args:
             if sep == tk.USING:
@@ -39,7 +39,7 @@ class Formatter(object):
                 self._print_value(next(args))
             newline = sep not in (tk.TAB, tk.SPC, b',', b';')
         if newline:
-            if self._screen and self._screen.overflow:
+            if self._console and self._console.overflow:
                 self._output.write_line()
             self._output.write_line()
 
@@ -78,7 +78,7 @@ class Formatter(object):
             self._output.write(b' ' * (pos-self._output.col), can_break=False)
 
     def _print_using(self, args):
-        """PRINT USING clause: Write expressions to screen or file using a formatting string."""
+        """PRINT USING clause: Write expressions to console or file using a formatting string."""
         format_expr = values.next_string(args)
         if format_expr == b'':
             raise error.BASICError(error.IFC)
@@ -115,6 +115,7 @@ class Formatter(object):
                             else:
                                 self._output.write(fors.read(1))
                             continue
+                    format_chars = True
                     value = next(args)
                     if value is None:
                         newline = False
@@ -122,8 +123,9 @@ class Formatter(object):
                     if start_cycle:
                         self._output.write(initial_literal)
                         start_cycle = False
-                        format_chars = True
                     self._output.write(format_field.format(value))
+            # consume any remaining arguments / finish parser
+            list(args)
         except StopIteration:
             pass
         if not format_chars:
@@ -264,7 +266,8 @@ class NumberField(object):
         valstr += b'$' if has_dollar else b''
         # format to string
         if b'^' in tokens:
-            valstr += value.to_str_scientific(digits_before, decimals, force_dot, comma)
+            # don't group thousands in scientific notation
+            valstr += value.to_str_scientific(digits_before, decimals, force_dot)
         else:
             valstr += value.to_str_fixed(decimals, force_dot, comma)
         # trailing signs, if any
