@@ -2,7 +2,7 @@
 PC-BASIC - ports.py
 Serial port handling
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -14,7 +14,7 @@ import io
 from contextlib import contextmanager
 
 from ...compat import iteritems
-from ...compat import console, stdin, stdout
+from ...compat import console, stdio
 
 from .devicebase import safe_io
 
@@ -258,7 +258,7 @@ class COMDevice(Device):
         with safe_io(error.DEVICE_FAULT):
             self._check_open()
             self._serial.baudrate = speed
-            self._serial.parity = parity
+            self._serial.parity = parity.decode('ascii')
             self._serial.bytesize = bytesize
             self._serial.stopbits = stop
 
@@ -267,7 +267,7 @@ class COMDevice(Device):
         with safe_io(error.DEVICE_FAULT):
             self._check_open()
             return (
-                self._serial.baudrate, self._serial.parity,
+                self._serial.baudrate, self._serial.parity.encode('ascii'),
                 self._serial.bytesize, self._serial.stopbits
             )
 
@@ -424,7 +424,7 @@ class SerialStdIO(object):
         self._crlf = crlf
         # dummy parameters
         self.baudrate = 300
-        self.parity = b'E'
+        self.parity = 'E'
         self.bytesize = 8
         self.stopbits = 2
         # dummy output pins
@@ -447,7 +447,7 @@ class SerialStdIO(object):
         # note that kbhit assumes keyboard
         # so won't work with redirects on Windows
         while console.key_pressed() and len(s) < num:
-            c = stdin.buffer.read(1)
+            c = stdio.stdin.buffer.read(1)
             if self._crlf and c == b'\n':
                 c = b'\r'
             if c:
@@ -458,8 +458,8 @@ class SerialStdIO(object):
         """Write to stdout."""
         if self._crlf:
             s = s.replace(b'\r', b'\n')
-        stdout.buffer.write(s)
-        stdout.buffer.flush()
+        stdio.stdout.buffer.write(s)
+        stdio.stdout.buffer.flush()
 
     @property
     def in_waiting(self):
