@@ -2,7 +2,7 @@
 PC-BASIC - eventcycle.py
 Event queue handling
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -55,23 +55,6 @@ HOME_KEY_REPLACEMENTS_EASCII = {
 ###############################################################################
 # queues
 
-def save_queue(q):
-    """Get list of queue tasks."""
-    qlist = []
-    while True:
-        try:
-            qlist.append(q.get(False))
-            q.task_done()
-        except queue.Empty:
-            break
-    return qlist
-
-def load_queue(q, qlist):
-    """Initialise queue from list of tasks."""
-    for item in qlist:
-        q.put(item)
-
-
 class NullQueue(object):
     """Dummy implementation of Queue interface."""
     def __init__(self, maxsize=0):
@@ -99,14 +82,15 @@ class EventQueues(object):
     """Manage interface queues."""
 
     tick = 0.006
-    max_video_qsize = 500
-    max_audio_qsize = 20
+    max_video_qsize = 200
+    #max_audio_qsize = 20
 
-    def __init__(self, values, ctrl_c_is_break, inputs=None, video=None, audio=None):
+    def __init__(self, ctrl_c_is_break, inputs=None, video=None, audio=None):
         """Initialise; default is NullQueues."""
-        self._values = values
         # input signal handlers
         self._handlers = []
+        # basic event handlers
+        self._basic_handlers = ()
         # pause-key halts everything until another keypress
         self._pause = False
         # treat ctrl+c as break interrupt
@@ -138,26 +122,45 @@ class EventQueues(object):
         """Add an input handler."""
         self._handlers.append(handler)
 
+    def set_basic_event_handlers(self, event_check_input):
+        """Set the handlers for BASIC events."""
+        self._basic_handlers = tuple(event_check_input)
+
     def wait(self):
         """Wait and check events."""
         time.sleep(self.tick)
         self.check_events()
 
-    def check_events(self, event_check_input=()):
+    def check_events(self):
         """Main event cycle."""
-        # check input first to avoid hang if the interface plugin has crashed
-        # and we have put a lot of work on the queue
-        # this works because Interface will send KEYB_QUIT on termination
-        self._check_input(event_check_input)
-        # avoid screen lockups if video queue fills up
+        # sleep(0) is needed for responsiveness, e.g. event trapping in programs with tight loops
+        # i.e. 100 goto 100 with event traps active) - needed to allow the input queue to fill
+        # this also allows the screen to update between statements
+        # it does slow the interpreter down by about 20% in FOR loops
+        # note that we always have an input queue, either for the interface of for iostreams
+        time.sleep(0)
+        # bizarrely, we need sleep(0) twice. I don't know why.
+        # note that multiple sleep(0) calls doen't seem to cause more slowdown than just one.
+        time.sleep(0)
+        time.sleep(0)
+        # what I think is happening here is that the sdl2 interface thread,
+        # in its loop to process a single queue item, calls C library functions
+        # which do not need the GIL. so it releases it. this allows the engine thread to pick up
+        # and produce more work for the interface thread. sleep(0) does not wait but it does release
+        # the GIL back, so we need a few sleep(0) calls to allow the interface to get through
+        # individual items.
+        # this would not be a problem if the interface thread did not need the GIL at all
+        # (perhaps with numba, nuitka, cython, pypy or jython)
+        # but note that the video queue is a Python object so may require the GIL
+        # or if it held the GIL for a full cycle
+        # wait to for the queue to drain if it excceds a threshold value
+        # this allows the interface to catch up with video updates
         if self.video.qsize() > self.max_video_qsize:
-            # note that this really slows down screen writing
-            # because it triggers a sleep() in the video backend
-            self.video.join()
-        if self.audio.qsize() > self.max_audio_qsize:
-            self.audio.join()
+            while self.video.qsize():
+                time.sleep(self.tick)
+        self._check_input()
 
-    def _check_input(self, event_check_input):
+    def _check_input(self):
         """Handle input events."""
         while True:
             # pop input queues
@@ -165,10 +168,11 @@ class EventQueues(object):
                 signal = self.inputs.get(False)
             except queue.Empty:
                 if self._pause:
+                    time.sleep(self.tick)
                     continue
                 else:
                     # we still need to handle basic events: not all are inputs
-                    for e in event_check_input:
+                    for e in self._basic_handlers:
                         e.check_input(signals.Event(None))
                     break
             self.inputs.task_done()
@@ -176,21 +180,21 @@ class EventQueues(object):
             self._replace_inputs(signal)
             # handle input events
             for handle_input in (
-                        [self._handle_non_trappable_interrupts] +
-                        [e.check_input for e in event_check_input] +
-                        [self._handle_trappable_interrupts] +
-                        [e.check_input for e in self._handlers]):
+                    [self._handle_non_trappable_interrupts] +
+                    [e.check_input for e in self._basic_handlers] +
+                    [self._handle_trappable_interrupts] +
+                    [e.check_input for e in self._handlers]
+                ):
                 if handle_input(signal):
                     break
 
     def _handle_non_trappable_interrupts(self, signal):
         """Handle non-trappable interrupts (before BASIC events)."""
         # process input events
-        if signal.event_type == signals.KEYB_QUIT:
+        if signal.event_type == signals.QUIT:
             raise error.Exit()
         # exit pause mode on keyboard hit; swallow key
-        elif signal.event_type in (
-                    signals.KEYB_DOWN, signals.STREAM_CHAR, signals.CLIP_PASTE):
+        elif signal.event_type in (signals.KEYB_DOWN, signals.STREAM_CHAR, signals.CLIP_PASTE):
             if self._pause:
                 self._pause = False
                 return True
@@ -201,8 +205,7 @@ class EventQueues(object):
         # handle special key combinations
         if signal.event_type == signals.KEYB_DOWN:
             c, scan, mod = signal.params
-            if (scan == scancode.DELETE and
-                    scancode.CTRL in mod and scancode.ALT in mod):
+            if (scan == scancode.DELETE and scancode.CTRL in mod and scancode.ALT in mod):
                 # ctrl-alt-del: if not captured by the OS, reset the emulator
                 # meaning exit and delete state. This is useful on android.
                 raise error.Reset()
