@@ -1,8 +1,8 @@
 """
-PC-BASIC - keyboard.py
-Keyboard and input stream handling
+PC-BASIC - inputs.keyboard
+Keyboard handling
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -18,7 +18,7 @@ from ..base.eascii import as_bytes as ea
 from ..base.eascii import as_unicode as uea
 
 
-# bit flags for modifier keys
+# bit flags for modifier keys - these are exposed via peek(1047) in low memory
 # sticky modifiers
 TOGGLE = {
     scancode.INSERT: 0x80, scancode.CAPSLOCK: 0x40,
@@ -172,22 +172,22 @@ class KeyboardBuffer(object):
 ###############################################################################
 # keyboard operations
 
-def _split_eascii(cp_s):
-    """Split a string of e-ascii/codepage into keystrokes."""
-    d = b''
-    for c in iterchar(cp_s):
-        if d or c != b'\0':
-            yield d + c
-            d = b''
-        elif c == b'\0':
+def _iter_keystrokes(ueascii_seq):
+    """Iterate over e-ascii/unicode sequences."""
+    out = u''
+    for char in ueascii_seq:
+        if out or char != u'\0':
+            yield out + char
+            out = u''
+        elif char == u'\0':
             # eascii code is \0 plus one char
-            d = c
+            out = char
 
 
 class Keyboard(object):
     """Keyboard handling."""
 
-    def __init__(self, queues, values, codepage, keystring, check_full):
+    def __init__(self, queues, values, codepage, check_full):
         """Initilise keyboard state."""
         self._values = values
         # needed for wait() in wait_char()
@@ -205,9 +205,6 @@ class Keyboard(object):
         self._ignore_caps = True
         # pre-inserted keystrings
         self._codepage = codepage
-        with self.buf.ignore_limit():
-            for ea_char in _split_eascii(self._codepage.str_from_unicode(keystring)):
-                self.buf.append(ea_char, None)
         # stream buffer
         self._stream_buffer = deque()
         # redirected input stream has closed
@@ -269,7 +266,7 @@ class Keyboard(object):
                 and not self._ignore_caps and len(c) == 1
             ):
             c = c.swapcase()
-        self.buf.append(self._codepage.from_unicode(c), scan)
+        self.buf.append(self._codepage.unicode_to_bytes(c), scan)
 
     def _key_up(self, scan):
         """Insert a key-up event."""
@@ -290,8 +287,18 @@ class Keyboard(object):
 
     def _stream_chars(self, us):
         """Insert eascii/unicode string into stream buffer."""
-        for ea_char in _split_eascii(self._codepage.str_from_unicode(us)):
-            self._stream_buffer.append(ea_char)
+        for ea_char in _iter_keystrokes(us):
+            ea_cp = self._codepage.unicode_to_bytes(ea_char)
+            # don't append empty strings arising from unknown unicode
+            # as this will be interpreted as end of stream, terminating the interpreter
+            if ea_cp:
+                self._stream_buffer.append(ea_cp)
+
+    def inject_keystrokes(self, keystring):
+        """Insert eascii/unicode string into keyboard buffer."""
+        with self.buf.ignore_limit():
+            for ea_char in _iter_keystrokes(keystring):
+                self.buf.append(self._codepage.unicode_to_bytes(ea_char), None)
 
     def _close_input(self):
         """Signal that input stream has closed."""
@@ -312,7 +319,7 @@ class Keyboard(object):
     # character retrieval
 
     def wait_char(self, keyboard_only=False):
-        """Wait for character, then return it but don't drop from queue."""
+        """Block until character appears in keyboard queue or stream."""
         # if input stream has closed, don't wait but return empty
         # which will tell the Editor to close
         # except if we're waiting for KYBD: input
@@ -323,7 +330,7 @@ class Keyboard(object):
             ):
             self._queues.wait()
 
-    def _read_byte(self, expand=True):
+    def _read_kybd_byte(self, expand=True):
         """Read one byte from keyboard buffer, expanding macros if required."""
         try:
             return self._expansion_vessel.pop(0)
@@ -344,10 +351,25 @@ class Keyboard(object):
     def inkey_(self, args):
         """INKEY$: read one byte from keyboard or stream; nonblocking."""
         list(args)
+        inkey = self.read_byte()
+        return self._values.new_string().from_str(inkey)
+
+    def read_byte(self):
+        """Read one byte from keyboard or stream; nonblocking."""
         # wait a tick to reduce load in loops
         self._queues.wait()
-        inkey = self._read_byte() or (self._stream_buffer.popleft() if self._stream_buffer else b'')
-        return self._values.new_string().from_str(inkey)
+        inkey = self._read_kybd_byte()
+        if not inkey and self._stream_buffer:
+            inkey = self._stream_buffer.popleft()
+        return inkey
+
+    def read_bytes_block(self, n):
+        """Read bytes from keyboard or stream; blocking."""
+        word = []
+        for _ in range(n):
+            self.wait_char(keyboard_only=False)
+            word.append(self.read_byte())
+        return b''.join(word)
 
     def peek_byte_kybd_file(self):
         """Peek from keyboard only; for KYBD: files; blocking."""
@@ -359,15 +381,15 @@ class Keyboard(object):
         word = []
         for _ in range(num):
             self.wait_char(keyboard_only=True)
-            word.append(self._read_byte(expand=False))
+            word.append(self._read_kybd_byte(expand=False))
         return word
 
     def get_fullchar(self, expand=True):
         """Read one (sbcs or dbcs) full character; nonblocking."""
-        c = self._read_byte(expand)
+        c = self._read_kybd_byte(expand)
         # insert dbcs chars from keyboard buffer two bytes at a time
         if (c in self._codepage.lead and self.buf.peek() in self._codepage.trail):
-            c += self._read_byte(expand)
+            c += self._read_kybd_byte(expand)
         if not c and self._stream_buffer:
             c = self._stream_buffer.popleft()
             if (c in self._codepage.lead and self._stream_buffer and
