@@ -2,7 +2,7 @@
 PC-BASIC - interpreter.py
 BASIC interpreter
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -18,15 +18,18 @@ from . import values
 class Interpreter(object):
     """BASIC interpreter."""
 
-    def __init__(self, queues, screen, files, sound,
-                values, memory, program, parser, basic_events):
+    def __init__(
+            self, queues, console, cursor, files, sound,
+            values, memory, program, parser, basic_events
+        ):
         """Initialise interpreter."""
         self._queues = queues
         self._basic_events = basic_events
         self._values = values
         self._memory = memory
         self._scalars = memory.scalars
-        self._screen = screen
+        self._console = console
+        self._cursor = cursor
         self._files = files
         self._sound = sound
         # program buffer
@@ -49,7 +52,7 @@ class Interpreter(object):
         self.set_pointer(False, 0)
         # interpreter is waiting for INPUT or LINE INPUT
         self.input_mode = False
-        # interpreter is executing a command (needs Screen)
+        # interpreter is executing a command (needs console)
         self.set_parse_mode(False)
         # additional operations on program step (debugging)
         self.step = lambda token: None
@@ -65,6 +68,15 @@ class Interpreter(object):
         """Unpickle."""
         self.__dict__.update(pickle_dict)
         self.step = lambda token: None
+        # go to start of statement on resume
+        ins = self.get_codestream()
+        ins.seek(self.current_statement)
+        if not self.parser.redo_on_break:
+            # parse line number or : at start of statement
+            if ins.read(1) in tk.END_LINE:
+                # line number marker, new statement
+                token = ins.read(4)
+            ins.skip_to(tk.END_STATEMENT)
 
     def _init_error_trapping(self):
         """Initialise error trapping."""
@@ -78,8 +90,10 @@ class Interpreter(object):
     def parse(self):
         """Parse from the current pointer in current codestream."""
         while True:
+            # update what basic events need to be handled
+            self._queues.set_basic_event_handlers(self._basic_events.enabled)
             # check input and BASIC events. may raise Break, Reset or Exit
-            self._queues.check_events(self._basic_events.enabled)
+            self._queues.check_events()
             try:
                 self.handle_basic_events()
                 ins = self.get_codestream()
@@ -101,7 +115,7 @@ class Interpreter(object):
                         return
                     if self.tron:
                         linenum = struct.unpack_from('<H', token, 2)
-                        self._screen.write(b'[%i]' % linenum)
+                        self._console.write(b'[%i]' % linenum)
                     self.step(token)
                 elif c not in (b':', tk.THEN, tk.ELSE, tk.GOTO):
                     # new statement or branch of an IF statement allowed, nothing else
@@ -112,14 +126,12 @@ class Interpreter(object):
 
     def loop(self):
         """Run commands until control returns to user."""
-        if not self._parse_mode:
+        if not self.parse_mode:
             return
-        self._screen.cursor.reset_visibility()
         try:
             # parse until break or end
             self.parse()
         except error.Break as e:
-            # ctrl-break stops foreground and background sound
             self._sound.stop_all_sound()
             self._handle_break(e)
         # move pointer to the start of direct line (for both on and off!)
@@ -129,28 +141,37 @@ class Interpreter(object):
 
     def set_parse_mode(self, on):
         """Enter or exit parse mode."""
-        self._parse_mode = on
-        self._screen.cursor.default_visible = not on
+        self.parse_mode = on
+        self._cursor.set_direct(not on)
 
     def _handle_break(self, e):
         """Handle a Break event."""
         # print ^C at current position
         if not self.input_mode and not e.stop:
-            self._screen.write(b'^C')
+            self._console.write(b'^C')
         # if we're in a program, save pointer
         pos = -1
         if self.run_mode:
             pos = self.current_statement
             if self.parser.redo_on_break:
-                self.stop = pos
+                self.stop_pos = pos
             else:
                 self._program.bytecode.skip_to(tk.END_STATEMENT)
-                self.stop = self._program.bytecode.tell()
-        self._screen.start_line()
-        self._screen.write(e.get_message(self._program.get_line_number(pos)))
-        self.set_parse_mode(False)
-        self.input_mode = False
+                self.stop_pos = self._program.bytecode.tell()
         self.parser.redo_on_break = False
+        if self.error_handle_mode:
+            e.trapped_error_num = self.error_num
+            # when the error is trapped, error position shifts to start of line
+            line = self._program.get_line_number(self.error_pos)
+            if 0 <= line <= 65535:
+                e.trapped_error_pos = self._program.line_numbers[line]
+            else:
+                e.trapped_error_pos = self.error_pos
+            #self.error_handle_mode = False
+        # ensure we can handle the break like an error
+        e.err = 0
+        e.pos = pos
+        raise e
 
     ###########################################################################
     # clear state
@@ -164,33 +185,33 @@ class Interpreter(object):
         # disable all event trapping (resets PEN to OFF too)
         self._basic_events.reset()
         # CLEAR also dumps for_next and while_wend stacks
-        self.clear_loop_stacks()
+        self.for_stack = []
+        self.while_stack = []
+        # reset stop/cont
+        self.stop_pos = None
         # reset the DATA pointer
         self.data_pos = 0
 
     def clear_stacks_and_pointers(self):
         """Initialise the stacks and pointers for a new program."""
-        # stop running if we were
-        self.set_pointer(False)
         # reset loop stacks
-        self.clear_stacks()
+        self._clear_stacks()
         # reset program pointer
         self._program_code.seek(0)
         # reset stop/cont
-        self.stop = None
+        self.stop_pos = None
         # reset data reader
         self.data_pos = 0
 
-    def clear_stacks(self):
+    def _clear_stacks(self):
         """Clear loop and jump stacks."""
+        # stop running if we were
+        self.set_pointer(False)
         self.gosub_stack = []
-        self.clear_loop_stacks()
-
-    def clear_loop_stacks(self):
-        """Clear loop stacks."""
         self.for_stack = []
         self.while_stack = []
 
+
     ###########################################################################
     # event and error handling
 
@@ -254,7 +275,10 @@ class Interpreter(object):
             self._files.lpt1_file.do_print()
         self.run_mode = new_runmode
         # events are active in run mode
-        self._basic_events.set_active(new_runmode)
+        if new_runmode:
+            self._queues.set_basic_event_handlers(self._basic_events.enabled)
+        else:
+            self._queues.set_basic_event_handlers([])
         # keep the sound engine on to avoid delays in run mode
         self._sound.persist(new_runmode)
         # suppress cassette messages in run mode
@@ -286,8 +310,10 @@ class Interpreter(object):
         """Execute jump for a GOSUB."""
         # set return position
         pos = self.get_codestream().tell()
+        # record run mode before executing self.jump, as that sets the runmode to True
+        run_mode = self.run_mode
         self.jump(jumpnum)
-        self.gosub_stack.append((pos, self.run_mode, handler))
+        self.gosub_stack.append((pos, run_mode, handler))
 
     def goto_(self, args):
         """GOTO: jump to line number."""
@@ -348,14 +374,14 @@ class Interpreter(object):
         i = -1
         for i, jumpnum in enumerate(args):
             if i == onvar-1:
+                # we counted the right number of commas, then didn't find a line number
+                if jumpnum is None:
+                    raise error.BASICError(error.STX)
                 if jump_type == tk.GOTO:
                     self.jump(jumpnum)
                 elif jump_type == tk.GOSUB:
                     self.jump_sub(jumpnum)
                 return
-        if i == onvar-2:
-            # missing jump *just where we need it* is an error
-            raise error.BASICError(error.STX)
 
     ###########################################################################
     # loops
@@ -550,8 +576,6 @@ class Interpreter(object):
                 value = self._values.from_str_at(word, address)
             else:
                 word = self._program_code.read_number()
-                if word is None:
-                    word = b''
                 value = self._values.from_repr(word, allow_nonnum=False)
                 # anything after the number is a syntax error, but assignment has taken place)
                 if (self._program_code.skip_blank() not in (tk.END_STATEMENT + (b',',))):
@@ -632,10 +656,10 @@ class Interpreter(object):
     def cont_(self, args):
         """CONT: continue STOPped or ENDed execution."""
         list(args)
-        if self.stop is None:
+        if self.stop_pos is None:
             raise error.BASICError(error.CANT_CONTINUE)
         else:
-            self.set_pointer(True, self.stop)
+            self.set_pointer(True, self.stop_pos)
         # IN GW-BASIC, weird things happen if you do GOSUB nn :PRINT "x"
         # and there's a STOP in the subroutine.
         # CONT then continues and the rest of the original line is executed, printing x
@@ -720,11 +744,10 @@ class Interpreter(object):
         new, old = self._program.explicit_lines(new, old)
         if step is not None and step < 1:
             raise error.BASICError(error.IFC)
-        old_to_new = self._program.renum(self._screen, new, old, step)
+        old_to_new = self._program.renum(self._console, new, old, step)
         # stop running if we were
-        self.set_pointer(False)
         # reset loop stacks
-        self.clear_stacks()
+        self._clear_stacks()
         # renumber error handler
         if self.on_error:
             self.on_error = old_to_new[self.on_error]
