@@ -2,7 +2,7 @@
 PC-BASIC - arrays.py
 Array variable management
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -23,8 +23,7 @@ class Arrays(object):
         self._memory = memory
         self._values = values
         self.clear()
-        # OPTION BASE is unset
-        self._base = None
+        self.clear_base()
 
     def __contains__(self, varname):
         """Check if a scalar has been defined."""
@@ -37,7 +36,11 @@ class Arrays(object):
     def __repr__(self):
         """Debugging representation of variable dictionary."""
         return '\n'.join(
-            '%s%s: %s' % (n, v, binascii.hexlify(bytes(self._buffers[n])))
+            '%s%s: %s' % (
+                n.decode('ascii'),
+                v,
+                binascii.hexlify(bytes(self._buffers[n])).decode('ascii')
+            )
             for n, v in iteritems(self._dims)
         )
 
@@ -45,7 +48,6 @@ class Arrays(object):
         """Clear arrays."""
         self._dims = {}
         self._buffers = {}
-        self._cache = {}
         self._array_memory = {}
         self.current = 0
 
@@ -58,12 +60,11 @@ class Arrays(object):
                 raise error.BASICError(error.IFC)
             dimensions = self._dims[name]
             record_len = 1 + max(3, len(name)) + 3 + 2*len(dimensions)
-            freed_bytes = self.array_len(dimensions) * values.size_bytes(name) + record_len
+            freed_bytes = self._buffer_size(name, dimensions) + record_len
             erased_name_ptr, _ = self._array_memory[name]
             # delete buffers
             del self._dims[name]
             del self._buffers[name]
-            del self._cache[name]
             del self._array_memory[name]
             # update memory model
             for name in self._array_memory:
@@ -71,6 +72,10 @@ class Arrays(object):
                 if name_ptr > erased_name_ptr:
                     self._array_memory[name] = name_ptr - freed_bytes, array_ptr - freed_bytes
             self.current -= freed_bytes
+        # if all arrays have been cleared and array base was set to 0 implicitly by DIM, unset it
+        # however, if array base was set explicitly by OPTION BASE, it remains set.
+        if not self._dims and self._base_set_by_dim:
+            self.clear_base()
 
     def index(self, index, dimensions):
         """Return the flat index for a given dimensioned index."""
@@ -82,18 +87,6 @@ class Arrays(object):
             area *= dimensions[i] + 1 - self._base
         return bigindex
 
-    def array_len(self, dimensions):
-        """Return the flat length for given dimensioned size."""
-        return self.index(dimensions, dimensions) + 1
-
-    def array_size_bytes(self, name):
-        """Return the byte size of an array, if it exists. Return 0 otherwise."""
-        try:
-            dimensions = self._dims[name]
-        except KeyError:
-            return 0
-        return self.array_len(dimensions) * values.size_bytes(name)
-
     def view_full_buffer(self, name):
         """Return a memoryview to a full array."""
         return memoryview(self._buffers[name])
@@ -102,14 +95,6 @@ class Arrays(object):
         """Return the dimensions of an array."""
         return self._dims[name]
 
-    def get_cache(self, name):
-        """Retrieve the sprite cache for the given array."""
-        return self._cache[name]
-
-    def set_cache(self, name, item):
-        """Store a sprite in the cache for a given array."""
-        self._cache[name] = item
-
     def dim_(self, args):
         """DIM: dimension arrays."""
         for a in args:
@@ -122,9 +107,13 @@ class Arrays(object):
         # first two bytes: chars of name or 0 if name is one byte long
         return 1 + max(3, len(name)) + 3 + 2*len(dimensions)
 
+    def flat_length(self, dimensions):
+        """Total number of elements."""
+        return self.index(dimensions, dimensions) + 1
+
     def _buffer_size(self, name, dimensions):
         """Calculate size of array buffer in bytes."""
-        return self.array_len(dimensions) * values.size_bytes(name)
+        return self.flat_length(dimensions) * values.size_bytes(name)
 
     def memory_size(self, name, dimensions):
         """Calculate size of array record and buffer in bytes."""
@@ -138,15 +127,17 @@ class Arrays(object):
         if not dimensions:
             # DIM A does nothing
             return
-        if self._base is None:
-            self._base = 0
         if name in self._dims:
             raise error.BASICError(error.DUPLICATE_DEFINITION)
-        for d in dimensions:
-            if d < 0:
-                raise error.BASICError(error.IFC)
-            elif d < self._base:
-                raise error.BASICError(error.SUBSCRIPT_OUT_OF_RANGE)
+        # a call that raises ifc does not implicitly set the array base
+        if any(_d < 0 for _d in dimensions):
+            raise error.BASICError(error.IFC)
+        # implicitly set array base
+        if self._base is None:
+            self._base = 0
+            self._base_set_by_dim = True
+        elif any(_d < self._base for _d in dimensions):
+            raise error.BASICError(error.SUBSCRIPT_OUT_OF_RANGE)
         # update memory model
         name_ptr = self.current
         record_len = self._record_size(name, dimensions)
@@ -158,7 +149,6 @@ class Arrays(object):
         self._array_memory[name] = (name_ptr, array_ptr)
         self._buffers[name] = bytearray(array_bytes)
         self._dims[name] = dimensions
-        self._cache[name] = None
 
     def check_dim(self, name, index):
         """
@@ -185,7 +175,10 @@ class Arrays(object):
 
     def clear_base(self):
         """Unset the array base."""
+        # OPTION BASE value. NONE: unset
         self._base = None
+        # OPTION BASE set by DIM rather than explicitly
+        self._base_set_by_dim = False
 
     def option_base_(self, args):
         """Set the array base to 0 or 1 (OPTION BASE). Raise error if already set."""
@@ -215,8 +208,7 @@ class Arrays(object):
             self._memory.strings.fix_temporaries()
         # copy value into array
         self.view_buffer(name, index)[:] = values.to_type(name[-1:], value).to_bytes()
-        # drop cache
-        self._cache[name] = None
+        # drop cache here
 
     def varptr(self, name, indices):
         """Retrieve the address of an array."""
@@ -247,36 +239,36 @@ class Arrays(object):
         """Retrieve data from data memory: array space """
         name_addr = -1
         arr_addr = -1
-        the_arr = None
         for name in self._array_memory:
             name_try, arr_try = self._array_memory[name]
             if name_try <= address and name_try > name_addr:
                 name_addr, arr_addr = name_try, arr_try
                 the_arr = name
-        if the_arr is None:
+                break
+        else: # pragma: no cover
             return -1
         var_current = self._memory.var_current()
+        dimensions = self._dims[the_arr]
         if address >= var_current + arr_addr:
             offset = address - arr_addr - var_current
-            if offset >= self.array_size_bytes(the_arr):
+            if offset >= self._buffer_size(the_arr, dimensions): # pragma: no cover
                 return -1
             byte_array = self._buffers[the_arr]
-            return byte_array[offset]
+            return ord(byte_array[offset:offset+1])
         else:
             offset = address - name_addr - var_current
             if offset < max(3, len(the_arr))+1:
                 return get_name_in_memory(the_arr, offset)
             else:
                 offset -= max(3, len(the_arr))+1
-                dimensions = self._dims[the_arr]
                 data_rep = struct.pack(
                     '<HB',
-                    self.array_size_bytes(the_arr) + 1 + 2*len(dimensions),
+                    self._buffer_size(the_arr, dimensions) + 1 + 2*len(dimensions),
                     len(dimensions)
                 )
                 for d in dimensions:
                     data_rep += struct.pack('<H', d + 1 - self._base)
-                return data_rep[offset]
+                return ord(data_rep[offset:offset+1])
 
     def get_strings(self):
         """Return a list of views of string array elements."""
@@ -298,7 +290,7 @@ class Arrays(object):
     def _from_list(self, python_list, name, index):
         """Convert Python list to BASIC array."""
         if not python_list:
-            return
+            raise ValueError('Array must not be empty.')
         if isinstance(python_list[0], list):
             for i, v in enumerate(python_list):
                 self._from_list(v, name, index+[i+(self._base or 0)])
@@ -308,23 +300,20 @@ class Arrays(object):
 
     def to_list(self, name):
         """Convert BASIC array to Python list."""
-        if name in self._dims:
-            indices = self._dims[name]
-            return self._to_list(name, [], indices)
-        else:
+        if name not in self._dims:
             return []
+        indices = self._dims[name]
+        return self._to_list(name, [], indices)
 
     def _to_list(self, name, index, remaining_dimensions):
         """Convert BASIC array to Python list."""
-        if not remaining_dimensions:
-            return []
-        elif len(remaining_dimensions) == 1:
+        if len(remaining_dimensions) == 1:
             return [
-                self.get(name, index+[i+(self._base or 0)]).to_value()
-                for i in range(remaining_dimensions[0])
+                self.get(name, index+[i]).to_value()
+                for i in range((self._base or 0), remaining_dimensions[0] + 1)
             ]
         else:
             return [
-                self._to_list(name, index+[i+(self._base or 0)], remaining_dimensions[1:])
-                for i in range(remaining_dimensions[0])
+                self._to_list(name, index+[i], remaining_dimensions[1:])
+                for i in range((self._base or 0), remaining_dimensions[0] + 1)
             ]
