@@ -2,27 +2,23 @@
 PC-BASIC - statements.py
 Statement parser
 
-(c) 2013--20186 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
 import logging
-import string
 import struct
 from functools import partial
 
+from ...compat import iterchar
 from ..base import error
 from ..base import tokens as tk
+from ..base.tokens import DIGITS, LETTERS
 from .. import values
 from . import expressions
 from . import userfunctions
 
 
-# bytes constants
-DIGITS = string.digits
-LETTERS = string.ascii_letters
-
-
 class Parser(object):
     """BASIC statement parser."""
 
@@ -76,7 +72,7 @@ class Parser(object):
             parse_args = stat_dict[selector]
         else:
             ins.seek(-len(c), 1)
-            if c in set(LETTERS):
+            if c in set(iterchar(LETTERS)):
                 # implicit LET
                 c = tk.LET
                 parse_args = self._simple[tk.LET]
@@ -333,10 +329,10 @@ class Parser(object):
             tk.BLOAD: session.all_memory.bload_,
             tk.SOUND: session.sound.sound_,
             tk.BEEP: session.sound.beep_,
-            tk.PSET: session.drawing.pset_,
-            tk.PRESET: session.drawing.preset_,
+            tk.PSET: session.graphics.pset_,
+            tk.PRESET: session.graphics.preset_,
             tk.SCREEN: session.display.screen_,
-            tk.LOCATE: session.screen.locate_,
+            tk.LOCATE: session.text_screen.locate_,
             tk.FILES: session.files.files_,
             tk.FIELD: session.files.field_,
             tk.NAME: session.files.name_,
@@ -346,10 +342,10 @@ class Parser(object):
             tk.CHAIN: session.chain_,
             tk.DATE: session.clock.date_,
             tk.TIME: session.clock.time_,
-            tk.PAINT: session.drawing.paint_,
+            tk.PAINT: session.graphics.paint_,
             tk.COM: session.basic_events.com_,
-            tk.CIRCLE: session.drawing.circle_,
-            tk.DRAW: session.drawing.draw_,
+            tk.CIRCLE: session.graphics.circle_,
+            tk.DRAW: session.graphics.draw_,
             tk.TIMER: session.basic_events.timer_,
             tk.IOCTL: session.files.ioctl_statement_,
             tk.CHDIR: session.files.chdir_,
@@ -357,7 +353,7 @@ class Parser(object):
             tk.RMDIR: session.files.rmdir_,
             tk.SHELL: session.shell_,
             tk.ENVIRON: session.environment.environ_statement_,
-            tk.WINDOW: session.drawing.window_,
+            tk.WINDOW: session.graphics.window_,
             tk.LCOPY: session.files.lcopy_,
             tk.PCOPY: session.display.pcopy_,
             tk.LOCK: session.files.lock_,
@@ -378,24 +374,24 @@ class Parser(object):
             tk.DEF + tk.USR: session.all_memory.def_usr_,
             tk.DEF: session.all_memory.def_seg_,
             tk.LINE + tk.INPUT: session.line_input_,
-            tk.LINE: session.drawing.line_,
-            tk.KEY + tk.ON: session.fkey_macros.key_,
-            tk.KEY + tk.OFF: session.fkey_macros.key_,
-            tk.KEY + tk.LIST: session.fkey_macros.key_,
+            tk.LINE: session.graphics.line_,
+            tk.KEY + tk.ON: session.console.key_,
+            tk.KEY + tk.OFF: session.console.key_,
+            tk.KEY + tk.LIST: session.console.key_,
             tk.KEY + b'(': session.basic_events.key_,
             tk.KEY: session.key_,
-            tk.PUT + b'(': session.drawing.put_,
+            tk.PUT + b'(': session.graphics.put_,
             tk.PUT: session.files.put_,
-            tk.GET + b'(': session.drawing.get_,
+            tk.GET + b'(': session.graphics.get_,
             tk.GET: session.files.get_,
             tk.PLAY + tk.ON: session.basic_events.play_,
             tk.PLAY + tk.OFF: session.basic_events.play_,
             tk.PLAY + tk.STOP: session.basic_events.play_,
             tk.PLAY: session.sound.play_,
-            tk.VIEW + tk.PRINT: session.screen.view_print_,
-            tk.VIEW: session.drawing.view_,
-            tk.PALETTE + tk.USING: session.display.palette.palette_using_,
-            tk.PALETTE: session.display.palette.palette_,
+            tk.VIEW + tk.PRINT: session.text_screen.view_print_,
+            tk.VIEW: session.graphics.view_,
+            tk.PALETTE + tk.USING: session.display.palette_using_,
+            tk.PALETTE: session.display.palette_,
             tk.STRIG + tk.ON: session.stick.strig_statement_,
             tk.STRIG + tk.OFF: session.stick.strig_statement_,
             tk.STRIG: session.basic_events.strig_,
@@ -467,7 +463,7 @@ class Parser(object):
         """Parse nothing."""
         # e.g. TRON LAH raises error but TRON will have been executed
         return
-        yield
+        yield # pragma: no cover
 
     def _parse_end(self, ins):
         """Parse end-of-statement before executing argumentless statement."""
@@ -475,19 +471,19 @@ class Parser(object):
         ins.require_end()
         # empty generator
         return
-        yield
+        yield # pragma: no cover
 
     def _skip_line(self, ins):
         """Ignore the rest of the line."""
         ins.skip_to(tk.END_LINE)
         return
-        yield
+        yield # pragma: no cover
 
     def _skip_statement(self, ins):
         """Ignore rest of statement."""
         ins.skip_to(tk.END_STATEMENT)
         return
-        yield
+        yield # pragma: no cover
 
     ###########################################################################
     # single argument
@@ -585,10 +581,10 @@ class Parser(object):
 
     def _parse_on_event(self, ins):
         """Helper function for ON event trap definitions."""
+        # token is known to be in (tk.PEN, tk.KEY, tk.TIMER, tk.PLAY, tk.COM, tk.STRIG)
+        # before we call this generator
         token = ins.read_keyword_token()
         yield token
-        if token not in (tk.PEN, tk.KEY, tk.TIMER, tk.PLAY, tk.COM, tk.STRIG):
-            raise error.BASICError(error.STX)
         if token != tk.PEN:
             yield self._parse_bracket(ins)
         else:
@@ -634,7 +630,7 @@ class Parser(object):
             # only look for args 3 and 4 if duration is > 0;
             # otherwise those args are a syntax error (on tandy)
             if (dur.sign() == 1) and ins.skip_blank_read_if((b',',)) and self._syntax in ('pcjr', 'tandy'):
-                yield self.parse_expression(ins)
+                yield self.parse_expression(ins, allow_empty=True)
                 if ins.skip_blank_read_if((b',',)):
                     yield self.parse_expression(ins)
                 else:
@@ -675,7 +671,7 @@ class Parser(object):
 
     def _parse_def_usr(self, ins):
         """Parse DEF USR syntax."""
-        ins.require_read((tk.USR))
+        ins.require_read((tk.USR,))
         yield ins.skip_blank_read_if(tk.DIGIT)
         ins.require_read((tk.O_EQ,))
         yield self.parse_expression(ins)
@@ -910,14 +906,14 @@ class Parser(object):
         if ins.skip_blank_read_if((tk.LOCK,), 2):
             yield self._parse_read_write(ins)
         else:
-            yield ins.skip_blank_read_if((tk.W_SHARED), 6)
+            yield ins.skip_blank_read_if((tk.W_SHARED,), 6)
         # AS file number clause
         ins.require_read((tk.W_AS,))
         ins.skip_blank_read_if((b'#',))
         yield self.parse_expression(ins)
         # LEN clause
         if ins.skip_blank_read_if((tk.LEN,), 2):
-            ins.require_read(tk.O_EQ)
+            ins.require_read((tk.O_EQ,))
             yield self.parse_expression(ins)
         else:
             yield None
@@ -1067,10 +1063,18 @@ class Parser(object):
             ins.require_read((tk.O_MINUS,))
             for c in self._parse_pair(ins):
                 yield c
+            if ins.skip_blank_read_if((b',',)):
+                fill_comma = True
+                fill = self.parse_expression(ins, allow_empty=True)
+                yield fill
+            else:
+                fill_comma = False
+                yield None
             if ins.skip_blank_read_if((b',',)):
                 yield self.parse_expression(ins)
-                ins.require_read((b',',))
-                yield self.parse_expression(ins)
+            else:
+                error.throw_if(fill_comma and not fill, error.MISSING_OPERAND)
+                yield None
 
     def _parse_line(self, ins):
         """Parse LINE syntax."""
@@ -1164,7 +1168,7 @@ class Parser(object):
 
     def _parse_def_fn(self, ins):
         """DEF FN: define a function."""
-        ins.require_read((tk.FN))
+        ins.require_read((tk.FN,))
         yield self.parse_name(ins)
 
     def _parse_var_list(self, ins):
@@ -1177,10 +1181,10 @@ class Parser(object):
     def _parse_deftype(self, ins):
         """Parse DEFSTR/DEFINT/DEFSNG/DEFDBL syntax."""
         while True:
-            start = ins.require_read(LETTERS)
+            start = ins.require_read(tuple(iterchar(LETTERS)))
             stop = None
             if ins.skip_blank_read_if((tk.O_MINUS,)):
-                stop = ins.require_read(LETTERS)
+                stop = ins.require_read(tuple(iterchar(LETTERS)))
             yield start, stop
             if not ins.skip_blank_read_if((b',',)):
                 break
@@ -1277,7 +1281,7 @@ class Parser(object):
         yield self._parse_variable(ins)
 
     ###########################################################################
-    # console and editor statements
+    # console / text screen statements
 
     def _parse_key_macro(self, ins):
         """Parse KEY ON/OFF/LIST syntax."""
@@ -1378,7 +1382,7 @@ class Parser(object):
             yield self.parse_expression(ins)
         else:
             yield None
-            if ins.peek() in set(DIGITS) | set(tk.NUMBER):
+            if ins.peek() in set(iterchar(DIGITS)) | set(tk.NUMBER):
                 expr = self.expression_parser.read_number_literal(ins)
             else:
                 expr = self.parse_expression(ins)
@@ -1473,8 +1477,6 @@ class Parser(object):
         yield ins.require_read((tk.GOTO, tk.GOSUB))
         while True:
             num = self._parse_optional_jumpnum(ins)
-            if num is None:
-                break
             yield num
             if not ins.skip_blank_read_if((b',',)):
                 break
