@@ -2,7 +2,7 @@
 PC-BASIC - numbers.py
 Integer and Floating Point values
 
-(c) 2013--2018 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3.
 """
 
@@ -26,14 +26,12 @@ import binascii
 import struct
 import math
 
+from ...compat import iterchar, int2byte
+
 from ..base import tokens as tk
 from ..base import error
 
 
-# mark bytes conversion explicitly
-int2byte = chr
-
-
 # for to_str
 # for numbers, tab and LF are whitespace
 BLANKS = b' \t\n'
@@ -58,17 +56,17 @@ class Value(object):
         self._buffer = memoryview(buffer)
         self._values = values
 
-    def __str__(self):
+    def __repr__(self):
         """String representation for debugging."""
+        sigil_repr = self.sigil.decode('ascii')
+        bytes_repr = binascii.hexlify(self.to_bytes()).decode('ascii')
         try:
-            return b'%s[%s %s]' % (
-                self.sigil, binascii.hexlify(bytes(self.to_bytes())), repr(self.to_value())
+            return '%s[%s %r]' % (
+                sigil_repr, bytes_repr,
+                self.to_value()
             )
         except Exception:
-            return b'%s[%s <detached>]' % (self.sigil, binascii.hexlify(bytes(self.to_bytes())))
-
-    __repr__ = __str__
-
+            return '%s[%s <detached>]' % (sigil_repr, bytes_repr)
 
     def __getstate__(self):
         # can't pickle memoryview
@@ -160,11 +158,16 @@ class Integer(Number):
 
     def is_negative(self):
         """Value is negative."""
-        return (ord(self._buffer[-1]) & 0x80) != 0
+        return (bytearray(self._buffer)[-1] & 0x80) != 0
 
     def sign(self):
         """Sign of value."""
-        return -1 if (ord(self._buffer[-1]) & 0x80) != 0 else (0 if self._buffer == b'\0\0' else 1)
+        if (bytearray(self._buffer)[-1] & 0x80) != 0:
+            return -1
+        elif bytearray(self._buffer) == bytearray(b'\0\0'):
+            return 0
+        else:
+            return 1
 
     def to_int(self, unsigned=False):
         """Return value as Python int."""
@@ -210,20 +213,16 @@ class Integer(Number):
 
     def to_token(self):
         """Return signed value as integer token."""
-        if self._buffer[1] == b'\0':
-            byte = ord(self._buffer[0])
+        if bytearray(self._buffer)[1] == 0:
+            byte = bytearray(self._buffer)[0]
             # although there is a one-byte token for '10', we don't write it.
             if byte < 10:
                 return int2byte(ord(tk.C_0) + byte)
             else:
-                return tk.T_BYTE + self._buffer[0]
+                return tk.T_BYTE + int2byte(bytearray(self._buffer)[0])
         else:
             return tk.T_INT + self._buffer.tobytes()
 
-    def to_token_linenum(self):
-        """Return unsigned value as line number token."""
-        return tk.T_UINT + self._buffer.tobytes()
-
     def to_token_hex(self):
         """Return unsigned value as hex token."""
         return tk.T_HEX + self._buffer.tobytes()
@@ -234,13 +233,13 @@ class Integer(Number):
 
     def from_token(self, token):
         """Set value to signed or unsigned integer token."""
-        d = bytes(token)[0]
-        if d in (tk.T_OCT, tk.T_HEX, tk.T_INT, tk.T_UINT):
+        d = bytearray(token)[0]
+        if d in (ord(_c) for _c in (tk.T_OCT, tk.T_HEX, tk.T_INT, tk.T_UINT)):
             self._buffer[:] = token[-2:]
-        elif d == tk.T_BYTE:
-            self._buffer[:] = token[-1] + b'\0'
-        elif tk.C_0 <= d <= tk.C_10:
-            self._buffer[:] = int2byte(ord(d) - 0x11) + b'\0'
+        elif d == ord(tk.T_BYTE):
+            self._buffer[:] = token[-1:] + b'\0'
+        elif ord(tk.C_0) <= d <= ord(tk.C_10):
+            self._buffer[:] = int2byte(d - 0x11) + b'\0'
         else:
             raise ValueError('%s is not an Integer token.' % repr(token))
         return self
@@ -251,16 +250,16 @@ class Integer(Number):
         """Convert integer to str in octal representation."""
         if self.is_zero():
             return b'0'
-        return oct(self.to_int(unsigned=True))[1:]
+        return b'%o' % (self.to_int(unsigned=True),)
 
     def to_hex(self):
         """Convert integer to str in hex representation."""
-        return hex(self.to_int(unsigned=True))[2:].upper()
+        return b'%X' % (self.to_int(unsigned=True),)
 
     def to_str(self, leading_space, type_sign):
         """Convert integer to str in decimal representation."""
-        intstr = bytes(self.to_int())
-        if leading_space and intstr[0] != b'-':
+        intstr = b'%d' % self.to_int()
+        if leading_space and intstr[0:1] != b'-':
             return b' ' + intstr
         else:
             return intstr
@@ -268,7 +267,7 @@ class Integer(Number):
     def from_oct(self, oct_repr):
         """Convert str in octal representation to integer."""
         # oct representations may be interrupted by blanks
-        val = int(oct_repr.strip(BLANKS), 8) if oct_repr else 0
+        val = int(oct_repr.strip(BLANKS), 8) if oct_repr.strip(BLANKS) else 0
         return self.from_int(val, unsigned=True)
 
     def from_hex(self, hex_repr):
@@ -279,14 +278,16 @@ class Integer(Number):
 
     def from_str(self, dec_repr):
         """Convert str in decimal representation to integer."""
-        return self.from_int(int(dec_repr.strip(BLANKS)))
+        valstr = dec_repr.strip(BLANKS)
+        if set(valstr) - set(tk.DIGITS):
+            raise ValueError(
+                'string %r contains non-digits, cannot convert to Integer.' %
+                (valstr,)
+            )
+        return self.from_int(int(valstr))
 
     # operations
 
-    def iround(self):
-        """Round in-place (no-op)."""
-        return self
-
     def itrunc(self):
         """Truncate towards zero in-place (no-op)."""
         return self
@@ -299,33 +300,38 @@ class Integer(Number):
         """Negate in-place."""
         if self._buffer == b'\x00\x80':
             raise error.BASICError(error.OVERFLOW)
-        lsb = (ord(self._buffer[0]) ^ 0xff) + 1
-        msb = ord(self._buffer[1]) ^ 0xff
+        lsb = (bytearray(self._buffer)[0] ^ 0xff) + 1
+        msb = bytearray(self._buffer)[1] ^ 0xff
         # apply carry
         if lsb > 0xff:
             lsb -= 0x100
             msb += 1
         # ignore overflow, since -0 == 0
-        self._buffer[:] = int2byte(lsb) + int2byte(msb & 0xff)
+        self._buffer[:] = bytearray([lsb, msb & 0xff])
         return self
 
     def iabs(self):
         """Absolute value in-place."""
-        if (ord(self._buffer[-1]) & 0x80):
+        if bytearray(self._buffer)[-1] & 0x80:
             return self.ineg()
+        return self
 
     def iadd(self, rhs):
         """Add another Integer in-place."""
-        lsb = ord(self._buffer[0]) + ord(rhs._buffer[0])
-        msb = ord(self._buffer[1]) + ord(rhs._buffer[1])
+        lsb = bytearray(self._buffer)[0] + bytearray(rhs._buffer)[0]
+        msb = bytearray(self._buffer)[1] + bytearray(rhs._buffer)[1]
         # apply carry
         if lsb > 0xff:
             lsb -= 0x100
             msb += 1
         # overflow if signs were equal and have changed
-        if (ord(self._buffer[1]) > 0x7f) == (ord(rhs._buffer[1]) > 0x7f) != (msb > 0x7f):
+        if (
+                (bytearray(self._buffer)[1] > 0x7f) ==
+                (bytearray(rhs._buffer)[1] > 0x7f) !=
+                (msb > 0x7f)
+            ):
             raise error.BASICError(error.OVERFLOW)
-        self._buffer[:] = int2byte(lsb) + int2byte(msb & 0xff)
+        self._buffer[:] = bytearray([lsb, msb & 0xff])
         return self
 
     def isub(self, rhs):
@@ -378,17 +384,17 @@ class Integer(Number):
         if isinstance(rhs, Float):
             # upgrade to Float
             return rhs.new().from_integer(self).gt(rhs)
-        isneg = ord(self._buffer[-1]) & 0x80
-        if isneg != (ord(rhs._buffer[-1]) & 0x80):
+        isneg = bytearray(self._buffer)[-1] & 0x80
+        if isneg != (bytearray(rhs._buffer)[-1] & 0x80):
             return not(isneg)
         # compute the unsigned (not absolute!) >
-        lmsb = ord(self._buffer[1]) & 0x7f
-        rmsb = ord(rhs._buffer[1]) & 0x7f
+        lmsb = bytearray(self._buffer)[1] & 0x7f
+        rmsb = bytearray(rhs._buffer)[1] & 0x7f
         if lmsb > rmsb:
             return True
         elif lmsb < rmsb:
             return False
-        return ord(self._buffer[0]) > ord(rhs._buffer[0])
+        return bytearray(self._buffer)[0] > bytearray(rhs._buffer)[0]
 
     def eq(self, rhs):
         """Equals."""
@@ -414,17 +420,17 @@ class Float(Number):
 
     def is_zero(self):
         """Value is zero."""
-        return self._buffer[-1] == b'\0'
+        return bytearray(self._buffer)[-1] == 0
 
     def is_negative(self):
         """Value is negative."""
-        return self._buffer[-2] >= b'\x80'
+        return bytearray(self._buffer)[-2] >= 0x80
 
     def sign(self):
         """Sign of value."""
-        if self._buffer[-1] == b'\0':
+        if bytearray(self._buffer)[-1] == 0:
             return 0
-        elif (ord(self._buffer[-2]) & 0x80) != 0:
+        elif (bytearray(self._buffer)[-2] & 0x80) != 0:
             return -1
         return 1
 
@@ -442,11 +448,11 @@ class Float(Number):
 
     def to_value(self):
         """Return value as Python float."""
-        exp = ord(self._buffer[-1]) - self._bias
-        if exp == -self._bias:
+        exp = bytearray(self._buffer)[-1] - self._bias
+        if exp == -self._bias:  # pylint: disable=invalid-unary-operand-type
             return 0.
         # unpack as unsigned long int
-        man = struct.unpack(self._intformat, bytearray(self._buffer[:-1]) + b'\0')[0]
+        man = struct.unpack(self._intformat, bytearray(self._buffer)[:-1] + b'\0')[0]
         # prepend assumed bit and apply sign
         if man & self._signmask:
             man = -man
@@ -470,7 +476,7 @@ class Float(Number):
         struct.pack_into(
             self._intformat, self._buffer, 0, man & (self._mask if neg else self._posmask)
         )
-        self._buffer[-1] = int2byte(exp)
+        self._buffer[-1:] = int2byte(exp)
         return self
 
 
@@ -484,7 +490,8 @@ class Float(Number):
             man += 0x80
         # note that -man >> 8 and -(man >> 8) are different
         # due to Python's rounding rules
-        return -(man >> 8) if neg else man >> 8
+        # python2: int() converts from type long if possible
+        return int(-(man >> 8) if neg else man >> 8)
 
     def from_int(self, in_int):
         """Set value to Python int."""
@@ -498,14 +505,15 @@ class Float(Number):
             struct.pack_into(
                 self._intformat, self._buffer, 0, man & (self._mask if neg else self._posmask)
             )
-            self._buffer[-1] = int2byte(exp)
+            self._buffer[-1:] = int2byte(exp)
         return self
 
     def to_int_truncate(self):
         """Truncate float to integer."""
         man, neg = self._to_int_den()
         # discard carry
-        return -(man >> 8) if neg else man >> 8
+        # python2: int() converts from type long if possible
+        return int(-(man >> 8) if neg else man >> 8)
 
     def mantissa(self):
         """Integer value of mantissa."""
@@ -516,18 +524,14 @@ class Float(Number):
 
     def ineg(self):
         """Negate in-place."""
-        self._buffer[-2] = int2byte(ord(self._buffer[-2]) ^ 0x80)
+        self._buffer[-2:-1] = int2byte(bytearray(self._buffer)[-2] ^ 0x80)
         return self
 
     def iabs(self):
         """Absolute value in-place."""
-        self._buffer[-2] = int2byte(ord(self._buffer[-2]) & 0x7F)
+        self._buffer[-2:-1] = int2byte(bytearray(self._buffer)[-2] & 0x7F)
         return self
 
-    def iround(self):
-        """Round in-place."""
-        return self.from_int(self.to_int())
-
     def itrunc(self):
         """Truncate towards zero in-place."""
         return self.from_int(self.to_int_truncate())
@@ -535,8 +539,10 @@ class Float(Number):
     def ifloor(self):
         """Truncate towards negative infinity in-place."""
         oldval = self.clone()
+        # we need to check negativity first, as itrunc() may take us to zero
+        was_negative = self.is_negative()
         self.itrunc()
-        if not self.eq(oldval) and self.is_negative():
+        if not self.eq(oldval) and was_negative:
             self.isub(self.new().from_bytes(self._one))
         return self
 
@@ -637,35 +643,36 @@ class Float(Number):
             sign = b'-'
         elif leading_space:
             sign = b' '
-        mantissa, exp10 = self.to_decimal()
-        digitstr = _get_digits(mantissa, self.digits, remove_trailing=True)
+        mantissa, exp10 = self.to_decimal(self.digits)
+        digitstr = _get_digits(mantissa, self.digits).rstrip(b'0')
         # exponent for scientific notation
         exp10 += self.digits - 1
         if exp10 > self.digits - 1 or len(digitstr)-exp10 > self.digits + 1:
             # use scientific notation
             # this is self.to_str_scientific(
-            #     n_before=1, n_decimals=self.digits-1, force_dot=False, group_digits=False)
+            #     n_before=1, n_decimals=self.digits-1, force_dot=False)
             valstr = self._scientific_notation(digitstr, exp10, digits_to_dot=1, force_dot=False)
         else:
             # use decimal notation
             # usual: n_decimals = -exp10
             # this is self.to_str_fixed
-            #     (self, n_decimals=-exp10, force_dot=False, group_digits=False) but with type_sign
+            #     (self, n_decimals=-exp10, force_dot=False, group_thousands=False) but with type_sign
             valstr = self._decimal_notation(digitstr, exp10, type_sign, force_dot=False)
         return sign + valstr
 
-    def to_decimal(self, digits=None):
-        """Return value as mantissa and decimal exponent."""
-        if digits is None:
+    def to_decimal(self, digits):
+        """Return value as mantissa of length min(digits, self.digits) and decimal exponent."""
+        if digits >= self.digits:
             # we'd be better off storing these (and self._ten) in denormalised form
             lim_bot = self.new().from_bytes(self._lim_bot)
             lim_top = self.new().from_bytes(self._lim_top)
-            digits = self.digits
         elif digits > 0:
             lim_bot = self.new().from_int(10**(digits-1))._just_under()
             lim_top = self.new().from_int(10**digits)._just_under()
         else:
-            return 0, 0
+            # use values for digits == 0 also if digits < 0
+            lim_bot = self.new().from_int(0)
+            lim_top = self.new().from_int(1)._just_under()
         tden = lim_top._denormalise()
         bden = lim_bot._denormalise()
         exp10 = 0
@@ -683,7 +690,8 @@ class Float(Number):
         # round to int
         exp, man, neg = den
         exp -= self._bias
-        if exp > 0:
+        if exp > 0: # pragma: no cover
+            # I don't think this happens
             man <<= exp
         else:
             man >>= -exp
@@ -705,32 +713,36 @@ class Float(Number):
             exp10 -= 1
         return self._normalise(*den)
 
-    def to_str_scientific(self, n_before, n_decimals, force_dot, group_digits):
-        """Put a float in scientific format."""
+    def to_str_scientific(self, digits_before_radix, digits_after_radix, always_show_radix):
+        """Put a positive float in scientific format."""
         if self.is_zero():
-            if force_dot:
-                return b''.join((b'.', (b'0' * n_decimals), self.exp_sign, b'+00'))
+            if always_show_radix:
+                return b''.join((b'.', b'0' * digits_after_radix, self.exp_sign, b'+00'))
             # single/double difference: this matches GW output. odd, odd, odd
             if self.exp_sign == b'E':
                 return b'E+00'
             return b'0D+00'
-        n_work = min(self.digits, n_before + n_decimals)
-        # special case when work_digits == 0, see also below
-        # setting to 0 results in incorrect rounding (why?)
-        mantissa, exp10 = self.to_decimal(1 if n_work == 0 else n_work)
-        digitstr = _get_digits(mantissa, n_work, remove_trailing=True)
-        # append zeros if necessary
-        digitstr = digitstr.ljust(n_decimals + n_before, b'0')
-        # this is just to reproduce GW results for no digits:
-        # e.g. PRINT USING "#^^^^";1 gives " E+01" not " E+00"
-        if n_work == 0:
-            exp10 += 1
-        exp10 += n_before + n_decimals - 1
+        digits_requested = digits_before_radix + digits_after_radix
+        digits_precision = self.digits
+        work_digits = min(digits_precision, digits_requested)
+        # convert to mantissa of length work_digits and exponent
+        mantissa, exponent = self.to_decimal(work_digits)
+        # radix position in the digit string
+        radix_position = exponent + work_digits
+        digitstr = _get_digits(mantissa, work_digits)
+        # ensure digitstr has the exact number of digits we're going to show
+        # when work_digits == 0, e.g. PRINT USING "#^^^^";1 gives " E+01" not " E+00"
+        # we get this because the digitstr is empty
+        digitstr = digitstr.ljust(digits_requested, b'0')[:digits_requested]
         return self._scientific_notation(
-            digitstr, exp10, digits_to_dot=n_before, force_dot=force_dot, group_digits=group_digits
+            digitstr,
+            # here, exp10 is the exponent if the radix is put after the first digit
+            exp10=radix_position-1,
+            digits_to_dot=digits_before_radix,
+            force_dot=always_show_radix
         )
 
-    def to_str_fixed(self, n_decimals, force_dot, group_digits):
+    def to_str_fixed(self, n_decimals, force_dot, group_thousands):
         """Put a float in fixed-point representation."""
         if self.is_zero():
             if force_dot:
@@ -739,7 +751,7 @@ class Float(Number):
                 return b'0'*n_decimals
             return b'0'
         # convert to integer_mantissa * 10**exponent
-        mantissa, exp10 = self.to_decimal()
+        mantissa, exp10 = self.to_decimal(self.digits)
         # -exp10 is the number of digits after the radix point
         n_after = -exp10
         # bring to decimal form of working precision
@@ -754,24 +766,23 @@ class Float(Number):
         # fill up with zeros to required number of figures
         digitstr = digitstr.ljust(n_decimals + n_before, b'0')
         return self._decimal_notation(
-            digitstr, n_before-1, type_sign=b'', force_dot=force_dot, group_digits=group_digits
+            digitstr, n_before-1, type_sign=b'',
+            force_dot=force_dot, group_thousands=group_thousands
         )
 
     # implementation: floating- and fixed-point decimal notations
 
-    def _group_digits(self, digitstr):
-        """Insert commas to group digits in a decimal number."""
+    def _group_thousands(self, digitstr):
+        """Insert commas to group digits by thousands in a decimal number."""
         first = len(digitstr) % 3
         chunks = [digitstr[i:i + 3] for i in range(first, len(digitstr), 3)]
         if first:
             chunks = [digitstr[:first]] + chunks
         return b','.join(chunks)
 
-    def _scientific_notation(self, digitstr, exp10, digits_to_dot, force_dot, group_digits=False):
+    def _scientific_notation(self, digitstr, exp10, digits_to_dot, force_dot):
         """Put digits in scientific E-notation."""
         valstr = digitstr[:digits_to_dot]
-        if group_digits:
-            valstr = self._group_digits(valstr)
         if len(digitstr) > digits_to_dot:
             after_str = digitstr[digits_to_dot:]
             valstr += b'.' + after_str
@@ -783,24 +794,24 @@ class Float(Number):
             valstr += b'-'
         else:
             valstr += b'+'
-        valstr += _get_digits(abs(exponent), n_digits=2, remove_trailing=False)
+        valstr += _get_digits(abs(exponent), min_digits=2)
         return valstr
 
-    def _decimal_notation(self, digitstr, exp10, type_sign, force_dot, group_digits=False):
+    def _decimal_notation(self, digitstr, exp10, type_sign, force_dot, group_thousands=False):
         """Put digits in decimal notation."""
         type_sign = b'' if not type_sign else self.sigil
         # digits to decimal point
         exp10 += 1
         if exp10 >= len(digitstr):
             valstr = digitstr + b'0'*(exp10-len(digitstr))
-            if group_digits:
-                valstr = self._group_digits(valstr)
+            if group_thousands:
+                valstr = self._group_thousands(valstr)
             if force_dot:
                 valstr += b'.'
         elif exp10 > 0:
             valstr = digitstr[:exp10]
-            if group_digits:
-                valstr = self._group_digits(valstr)
+            if group_thousands:
+                valstr = self._group_thousands(valstr)
             valstr += b'.' + digitstr[exp10:]
         else:
             valstr = b'.' + b'0'*(-exp10) + digitstr
@@ -878,9 +889,9 @@ class Float(Number):
 
     def _denormalise(self):
         """Denormalise to shifted mantissa, exp, sign."""
-        exp = ord(self._buffer[-1])
+        exp = bytearray(self._buffer)[-1]
         man = struct.unpack(
-                self._intformat, b'\0' + bytearray(self._buffer[:-1])
+                self._intformat, b'\0' + bytearray(self._buffer)[:-1]
             )[0] | self._den_mask
         neg = self.is_negative()
         return exp, man, neg
@@ -906,7 +917,7 @@ class Float(Number):
             self._intformat, self._buffer, 0, (man>>8) & (self._mask if neg else self._posmask)
         )
         if self._check_limits(exp, neg):
-            self._buffer[-1] = int2byte(exp)
+            self._buffer[-1:] = int2byte(exp)
         return self
 
     def _to_int_den(self):
@@ -948,18 +959,17 @@ class Float(Number):
 
     def _abs_gt(self, rhs):
         """Absolute values greater than."""
-        # don't compare zeroes
-        if self.is_zero():
+        # don't compare zeroes - failsafe, is not reached in code
+        if self.is_zero(): # pragma: no cover
             return False
         rhscopy = bytearray(rhs._buffer)
         # so long as the sign is the same ...
-        rhscopy[-2] &= (ord(self._buffer[-2]) | 0x7f)
+        rhscopy[-2] &= (bytearray(self._buffer)[-2] | 0x7f)
         # ... we can compare floats as if they were ints
-        for l, r in reversed(zip(self._buffer, rhscopy)):
-            # memoryview elements are str, bytearray elements are int
-            if ord(l) > r:
+        for l, r in reversed(list(zip(bytearray(self._buffer), bytearray(rhscopy)))):
+            if l > r:
                 return True
-            elif ord(l) < r:
+            elif l < r:
                 return False
         # equal
         return False
@@ -1081,7 +1091,7 @@ class Single(Float):
 
     def from_token(self, token):
         """Set value to Single token."""
-        if bytes(token)[0] != tk.T_SINGLE:
+        if bytearray(token)[0] != ord(tk.T_SINGLE):
             raise ValueError('%s is not a Single token.' % repr(token))
         self._buffer[:] = token[-4:]
         return self
@@ -1144,7 +1154,7 @@ class Double(Float):
 
     def from_token(self, token):
         """Set value to Single token."""
-        if bytes(token)[0] != tk.T_DOUBLE:
+        if bytearray(token)[0] != ord(tk.T_DOUBLE):
             raise ValueError('%s is not a Double token.' % repr(token))
         self._buffer[:] = token[-8:]
         return self
@@ -1178,7 +1188,7 @@ def str_to_decimal(s, allow_nonnum=True):
     found_exp_sign, exp_neg, neg = False, False, False
     exp10, exponent, mantissa, digits, zeros = 0, 0, 0, 0, 0
     is_double, is_single = False, False
-    for c in s:
+    for c in iterchar(s):
         # ignore whitespace throughout (x = 1   234  56  .5  means x=123456.5 in gw!)
         if c in BLANKS:
             continue
@@ -1207,7 +1217,7 @@ def str_to_decimal(s, allow_nonnum=True):
                     else:
                         zeros = 0
                 continue
-            elif c == '.':
+            elif c == b'.':
                 found_point = True
                 continue
             elif c.upper() in b'DE':
@@ -1249,10 +1259,6 @@ def str_to_decimal(s, allow_nonnum=True):
         is_double = True
     return is_double, -mantissa if neg else mantissa, exp10
 
-def _get_digits(mantissa, n_digits, remove_trailing):
+def _get_digits(mantissa, min_digits):
     """Get the digits for an int."""
-    digitstr = (b'%d' % abs(mantissa)).rjust(n_digits, b'0')
-    if remove_trailing:
-        return digitstr.rstrip(b'0')
-    else:
-        return digitstr
+    return (b'%d' % abs(mantissa)).rjust(min_digits, b'0')
