@@ -2,7 +2,7 @@
 PC-BASIC - interface.base
 Interface utility classes
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -18,14 +18,6 @@ NOKILL_MESSAGE = u'to exit type <CTRL+BREAK> <ESC> SYSTEM'
 class InitFailed(Exception):
     """Initialisation failed."""
 
-    def __init__(self, message=''):
-        """Set message."""
-        self._message = message
-
-    def __str__(self):
-        """Error description (overrides Python 2 default)."""
-        return self._message
-
 
 class PluginRegister(object):
     """Plugin register."""
