@@ -2,7 +2,7 @@
 PC-BASIC - interface.interface
 Interface class
 
-(c) 2013--2019 Rob Hagemans
+(c) 2013--2023 Rob Hagemans
 This file is released under the GNU GPL version 3 or later.
 """
 
@@ -25,13 +25,12 @@ DELAY = 12
 class Interface(object):
     """User interface for PC-BASIC session."""
 
-    def __init__(self, guard=None, try_interfaces=(), audio_override=None, wait=False, **kwargs):
+    def __init__(self, try_interfaces=(), audio_override=None, wait=False, **kwargs):
         """Initialise interface."""
         self._input_queue = queue.Queue()
         self._video_queue = queue.Queue()
         self._audio_queue = queue.Queue()
         self._wait = wait
-        self._guard = guard
         self._video, self._audio = None, None
         for video in try_interfaces:
             try:
@@ -50,7 +49,7 @@ class Interface(object):
             self._audio = audio_plugins[audio](self._audio_queue, **kwargs)
         except KeyError:
             # ignore if an interface has no audio, but not if an override doesn't exist
-            if audio_override and audio_override != 'none':
+            if audio_override and audio_override not in ('none', 'false'):
                 logging.error('Unknown audio plugin `%s`', audio)
         except InitFailed as e:
             logging.info('Could not initialise audio plugin `%s`: %s', audio, e)
@@ -62,9 +61,9 @@ class Interface(object):
         """Retrieve interface queues."""
         return self._input_queue, self._video_queue, self._audio_queue
 
-    def launch(self, target, **kwargs):
+    def launch(self, target, *args, **kwargs):
         """Start an interactive interpreter session."""
-        thread = threading.Thread(target=self._thread_runner, args=(target,), kwargs=kwargs)
+        thread = threading.Thread(target=self._thread_runner, args=(target,) + args, kwargs=kwargs)
         try:
             # launch the BASIC thread
             thread.start()
@@ -77,10 +76,10 @@ class Interface(object):
             self.quit_input()
             thread.join()
 
-    def _thread_runner(self, target, **kwargs):
+    def _thread_runner(self, target, *args, **kwargs):
         """Session runner."""
         try:
-            target(interface=self, guard=self._guard, **kwargs)
+            target(*args, **kwargs)
         finally:
             if self._wait:
                 self.pause(WAIT_MESSAGE)
@@ -97,22 +96,20 @@ class Interface(object):
                     if not self._audio.busy and not self._video.busy:
                         # nothing to do, come back later
                         self._video.sleep(DELAY)
-                    else:
-                        # tiny delay; significantly reduces cpu load when playing audio or blinking
-                        self._video.sleep(1)
 
     def pause(self, message):
         """Pause and wait for a key."""
         self._video_queue.put(signals.Event(signals.VIDEO_SET_CAPTION, (message,)))
-        self._video_queue.put(signals.Event(signals.VIDEO_SHOW_CURSOR, (False,)))
+        self._video_queue.put(signals.Event(signals.VIDEO_SHOW_CURSOR, (False, False)))
         while True:
             signal = self._input_queue.get()
-            if signal.event_type in (signals.KEYB_DOWN, signals.KEYB_QUIT):
-                break
+            if signal.event_type in (signals.KEYB_DOWN, signals.QUIT):
+                self._video_queue.put(signals.Event(signals.VIDEO_SET_CAPTION, (u'',)))
+                return signal
 
     def quit_input(self):
         """Send signal through the input queue to quit BASIC."""
-        self._input_queue.put(signals.Event(signals.KEYB_QUIT))
+        self._input_queue.put(signals.Event(signals.QUIT))
         # drain video queue (joined in other thread)
         while not self._video_queue.empty():
             try:
@@ -130,5 +127,5 @@ class Interface(object):
 
     def quit_output(self):
         """Send signal through the output queues to quit plugins."""
-        self._video_queue.put(signals.Event(signals.VIDEO_QUIT))
-        self._audio_queue.put(signals.Event(signals.AUDIO_QUIT))
+        self._video_queue.put(signals.Event(signals.QUIT))
+        self._audio_queue.put(signals.Event(signals.QUIT))
