@@ -12,6 +12,11 @@ import os
 import struct
 
 
+# deal with broken pipes
+
+def is_broken_pipe(e):
+    return isinstance(e, BrokenPipeError)
+
 # strings
 
 int2byte = struct.Struct(">B").pack
@@ -41,6 +46,10 @@ def iterchar(s):
     """Iterate over bytes, returning char."""
     return (s[_i:_i+1] for _i in range(len(s)))
 
+def iterbytes(s):
+    """Iterate over bytes/bytearray/memoryview, returning int."""
+    return s
+
 def iteritems(d, **kw):
     return iter(d.items(**kw))
 
